"""
Defines strategies for selecting and assigning colors to chart elements based on data attributes.
"""

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import ClassVar

import polars as pl
from polychromos.palette import HSLColorScale, HSLColorSequence, Palette


class ColorSelectorStrategy:
    """
    Base strategy for selecting data points to apply colors to.
    """

    type_id: ClassVar[str] = "All"

    def selector(
        self,
        df: pl.DataFrame,
        category_var: str,
        value_var: str,
    ) -> list[bool]:
        """
        Selects which entries in the DataFrame should be colored.

        :param df: The DataFrame containing the data.
        :type df: pl.DataFrame
        :param category_var: The name of the category variable.
        :type category_var: str
        :param value_var: The name of the value variable.
        :type value_var: str
        :return: A list of booleans indicating which rows are selected.
        :rtype: list[bool]
        """
        return [True] * df.height

    @classmethod
    def get_name(cls) -> str:
        """
        Gets the name of the color selector strategy.

        :return: Human-readable name.
        :rtype: str
        """
        return "All entries"

    @classmethod
    def get_description(cls) -> str:
        """
        Gets the description of the color selector strategy.

        :return: Human-readable description.
        :rtype: str
        """
        return (
            "Selects all entries. This should be used only in the first color rule, as it will "
            "override all previous rules' selection criteria."
        )

    @classmethod
    def get_icon(cls) -> str | None:
        """
        Gets an optional SVG icon for the color selector strategy.

        :return: An SVG string, or ``None`` if not set.
        :rtype: str | None
        """
        return None

    def explain(self) -> str:
        """
        Returns a human-readable participial phrase describing the selection criteria.

        :return: A phrase such as ``"selecting all entries"``.
        :rtype: str
        """
        return "selecting all entries"


class ColorSelectByCategoryName(ColorSelectorStrategy):
    """
    Selects data points based on category names.
    """

    type_id: ClassVar[str] = "ByCategoryName"

    def __init__(
        self,
        names: list[str],
    ) -> None:
        self._names = names

    def selector(
        self,
        df: pl.DataFrame,
        category_var: str,
        value_var: str,
    ) -> list[bool]:
        return [name in self._names for name in df[category_var]]

    @classmethod
    def get_name(cls) -> str:
        """Returns the name of this selector strategy."""
        return "By category names"

    @classmethod
    def get_description(cls) -> str:
        """Returns the description of this selector strategy."""
        return "Selects entries whose category names are in a given list."

    def explain(self) -> str:
        return f"selecting entries in categories {self._names}"


class ColorSelectValuesBelow(ColorSelectorStrategy):
    """
    Selects data points with values below a certain threshold.
    """

    type_id: ClassVar[str] = "ValuesBelow"

    def __init__(
        self,
        threshold: float,
    ) -> None:
        self._threshold = threshold

    def selector(
        self,
        df: pl.DataFrame,
        category_var: str,
        value_var: str,
    ) -> list[bool]:
        return [value <= self._threshold for value in df[value_var]]

    @classmethod
    def get_name(cls) -> str:
        """Returns the name of this selector strategy."""
        return "Values below threshold"

    @classmethod
    def get_description(cls) -> str:
        """Returns the description of this selector strategy."""
        return "Selects entries with values below a specified threshold."

    def explain(self) -> str:
        return f"selecting entries with a value below {self._threshold}"


class ColorSelectValuesAbove(ColorSelectorStrategy):
    """
    Selects data points with values above a certain threshold.
    """

    type_id: ClassVar[str] = "ValuesAbove"

    def __init__(
        self,
        threshold: float,
    ) -> None:
        self._threshold = threshold

    def selector(
        self,
        df: pl.DataFrame,
        category_var: str,
        value_var: str,
    ) -> list[bool]:
        return [value >= self._threshold for value in df[value_var]]

    @classmethod
    def get_name(cls) -> str:
        """Returns the name of this selector strategy."""
        return "Values above threshold"

    @classmethod
    def get_description(cls) -> str:
        """Returns the description of this selector strategy."""
        return "Selects entries with values above a specified threshold."

    def explain(self) -> str:
        return f"selecting entries with a value above {self._threshold}"


class ColorSelectMaxValue(ColorSelectorStrategy):
    """
    Selects data points with the maximum value.
    """

    type_id: ClassVar[str] = "MaxValue"

    def selector(
        self,
        df: pl.DataFrame,
        category_var: str,
        value_var: str,
    ) -> list[bool]:
        if df.height == 0:
            return []
        max_value = df[value_var].max()
        if max_value is None:
            return [False] * df.height
        return [value == max_value for value in df[value_var]]

    @classmethod
    def get_name(cls) -> str:
        """Returns the name of this selector strategy."""
        return "Maximum value"

    @classmethod
    def get_description(cls) -> str:
        """Returns the description of this selector strategy."""
        return "Selects entries with the maximum value."

    def explain(self) -> str:
        return "selecting the entry with the highest value"


class ColorSelectMinValue(ColorSelectorStrategy):
    """
    Selects data points with the minimum value.
    """

    type_id: ClassVar[str] = "MinValue"

    def selector(
        self,
        df: pl.DataFrame,
        category_var: str,
        value_var: str,
    ) -> list[bool]:
        if df.height == 0:
            return []
        min_value = df[value_var].min()
        if min_value is None:
            return [False] * df.height
        return [value == min_value for value in df[value_var]]

    @classmethod
    def get_name(cls) -> str:
        """Returns the name of this selector strategy."""
        return "Minimum value"

    @classmethod
    def get_description(cls) -> str:
        """Returns the description of this selector strategy."""
        return "Selects entries with the minimum value."

    def explain(self) -> str:
        return "selecting the entry with the lowest value"


class ColorSelectValuesBelowMean(ColorSelectorStrategy):
    """
    Selects data points with values below the total weighted mean.
    """

    type_id: ClassVar[str] = "ValuesBelowMean"

    def selector(
        self,
        df: pl.DataFrame,
        category_var: str,
        value_var: str,
    ) -> list[bool]:
        return [value <= df["__total_mean__"][i] for i, value in enumerate(df[value_var])]

    @classmethod
    def get_name(cls) -> str:
        """Returns the name of this selector strategy."""
        return "Values below mean"

    @classmethod
    def get_description(cls) -> str:
        """Returns the description of this selector strategy."""
        return "Selects entries with values below the total weighted mean."

    def explain(self) -> str:
        return "selecting entries with a value below the mean"


class ColorSelectValuesAboveMean(ColorSelectorStrategy):
    """
    Selects data points with values above the total weighted mean.
    """

    type_id: ClassVar[str] = "ValuesAboveMean"

    def selector(
        self,
        df: pl.DataFrame,
        category_var: str,
        value_var: str,
    ) -> list[bool]:
        return [value >= df["__total_mean__"][i] for i, value in enumerate(df[value_var])]

    @classmethod
    def get_name(cls) -> str:
        """Returns the name of this selector strategy."""
        return "Values above mean"

    @classmethod
    def get_description(cls) -> str:
        """Returns the description of this selector strategy."""
        return "Selects entries with values above the total weighted mean."

    def explain(self) -> str:
        return "selecting entries with a value above the mean"


class ColorAssignerStrategy(ABC):
    """
    Base strategy for assigning colors to data points.
    """

    type_id: ClassVar[str]

    def __init__(
        self,
        anchor_colors: HSLColorSequence,
    ) -> None:
        """
        Constructor method.

        :param anchor_colors: A list of colors used to generate the palette to assign colors from.
        :type anchor_colors: HSLColorSequence
        :raises ValueError: When no anchor colors are provided.
        """
        if len(anchor_colors) == 0:
            raise ValueError("No anchor colors provided")
        self._anchor_colors = anchor_colors

    @property
    def anchor_colors(self) -> HSLColorSequence:
        """
        The anchor colors used to generate the palette.

        :return: The sequence of anchor colors.
        :rtype: HSLColorSequence
        """
        return self._anchor_colors

    @property
    def continuous(self) -> bool:
        """
        Whether the color assigner is continuous (``True``) or discrete (``False``).
        """
        return False

    @abstractmethod
    def assign_colors(
        self,
        df: pl.DataFrame,
        category_var: str,
        value_var: str,
        group_var: str | None = None,
    ) -> HSLColorSequence:
        """
        Generates a palette of colors according to the strategy.

        :param df: The DataFrame containing the data.
        :type df: pl.DataFrame
        :param category_var: The name of the category variable.
        :type category_var: str
        :param value_var: The name of the value variable.
        :type value_var: str
        :param group_var: The name of the groups variable, if any. Defaults to ``None``.
        :type group_var: str | None, optional
        :return: A sequence of colors assigned to the data points.
        :rtype: HSLColorSequence
        """

    @classmethod
    @abstractmethod
    def get_name(cls) -> str:
        """
        Gets the name of the color assigner strategy.
        """

    @classmethod
    @abstractmethod
    def get_description(cls) -> str:
        """
        Gets the description of the color assigner strategy.
        """

    @classmethod
    def get_icon(cls) -> str | None:
        """
        Gets an optional SVG icon for the color assigner strategy.

        :return: An SVG string, or ``None`` if not set.
        :rtype: str | None
        """
        return None

    def _hex_anchors(self) -> list[str]:
        return [c.to_css_hex() for c in self._anchor_colors]

    @abstractmethod
    def explain(self) -> str:
        """
        Returns a human-readable imperative phrase describing the color assignment.

        :return: A phrase such as ``"colorize them each in a distinct color from ['#ff0000']"``.
        :rtype: str
        """


class DiscreteColorAssignerStrategy(ColorAssignerStrategy):
    """
    Strategy for assigning discrete colors to data points.
    """

    def __init__(
        self,
        anchor_colors: HSLColorSequence,
        shuffle: bool = False,
    ) -> None:
        """
        Constructor method.

        :param anchor_colors: A list of colors used to generate the palette to assign colors from.
        :type anchor_colors: HSLColorSequence
        :param shuffle: Whether to alternate colors or not. Defaults to ``False``.
        :type shuffle: bool, optional
        """
        super().__init__(anchor_colors=anchor_colors)
        self._shuffle = shuffle

    def _calculate_interpolated_sequence(
        self,
        n_colors: int,
    ) -> HSLColorSequence:
        if n_colors <= 1:
            return [self._anchor_colors[0]]

        sequence: HSLColorSequence
        if len(self._anchor_colors) == 1:
            sequence = self._anchor_colors * n_colors
        elif len(self._anchor_colors) == 2 and n_colors > 2:
            sequence = Palette.sequence_from_cylindrical_interpolation(
                self._anchor_colors[0],
                self._anchor_colors[1],
                n_colors,
                path_strategy=Palette.CylindricalInterpolationPath.SHORTEST,
            )
        else:
            scale: HSLColorScale = Palette.to_color_scale(self._anchor_colors)
            sequence = Palette.color_scale_to_color_sequence(scale, n_colors)

        if self._shuffle:
            sequence = Palette.alternate_colors(sequence)

        return sequence


class ColorAssignerByCategory(DiscreteColorAssignerStrategy):
    """
    Strategy for assigning colors based on category values.
    """

    type_id: ClassVar[str] = "ByCategory"

    def __init__(
        self,
        anchor_colors: HSLColorSequence,
        shuffle: bool = False,
    ) -> None:
        super().__init__(
            anchor_colors=anchor_colors,
            shuffle=shuffle,
        )

    def assign_colors(
        self,
        df: pl.DataFrame,
        category_var: str,
        value_var: str,
        group_var: str | None = None,
    ) -> HSLColorSequence:
        return self._calculate_interpolated_sequence(len(df[category_var]))

    @classmethod
    def get_name(cls) -> str:
        """Returns the name of this assigner strategy."""
        return "Assign by category"

    @classmethod
    def get_description(cls) -> str:
        """Returns the description of this assigner strategy."""
        return "Assigns colors based on the distinct category values."

    def explain(self) -> str:
        return f"colorize them each in a distinct color from {self._hex_anchors()}"


class ColorAssignerByGroup(DiscreteColorAssignerStrategy):
    """
    Strategy for assigning colors based on group values.
    """

    type_id: ClassVar[str] = "ByGroup"

    def __init__(
        self,
        anchor_colors: HSLColorSequence,
        shuffle: bool = False,
    ) -> None:
        super().__init__(
            anchor_colors=anchor_colors,
            shuffle=shuffle,
        )

    def assign_colors(
        self,
        df: pl.DataFrame,
        category_var: str,
        value_var: str,
        group_var: str | None = None,
    ) -> HSLColorSequence:
        if group_var is None:
            raise ValueError("Group variable must be provided for ColorAssignerByGroup")
        return self._calculate_interpolated_sequence(len(df[group_var].unique()))

    @classmethod
    def get_name(cls) -> str:
        """Returns the name of this assigner strategy."""
        return "Assign by group"

    @classmethod
    def get_description(cls) -> str:
        """Returns the description of this assigner strategy."""
        return "Assigns colors based on the distinct group values."

    def explain(self) -> str:
        return f"colorize them each in a distinct color by group from {self._hex_anchors()}"


class ColorAssignerByValue(ColorAssignerStrategy):
    """
    Strategy for assigning colors based on entry values.
    """

    type_id: ClassVar[str] = "ByValue"

    def __init__(
        self,
        anchor_colors: HSLColorSequence,
        range_min: float | None = None,
        range_max: float | None = None,
        range_center_at: float | None = None,
    ) -> None:
        super().__init__(anchor_colors=anchor_colors)
        self._range_min = range_min
        self._range_max = range_max
        self._range_center_at = range_center_at

    @property
    def continuous(self) -> bool:
        return True

    @property
    def range_min(self) -> float | None:
        """
        Gets the minimum value of the range for color assignment.
        """
        return self._range_min

    @property
    def range_max(self) -> float | None:
        """
        Gets the maximum value of the range for color assignment.
        """
        return self._range_max

    @property
    def range_center_at(self) -> float | None:
        """
        Gets the center value of the range for color assignment.
        """
        return self._range_center_at

    def assign_colors(
        self,
        df: pl.DataFrame,
        category_var: str,
        value_var: str,
        group_var: str | None = None,
    ) -> HSLColorSequence:
        scale: HSLColorScale = Palette.to_color_scale(self._anchor_colors if len(self._anchor_colors) > 1 else self._anchor_colors * 2)
        temp_min: float = self._range_min if self._range_min is not None else min(df[value_var])
        temp_max: float = self._range_max if self._range_max is not None else max(df[value_var])
        low_end: float
        high_end: float
        if self._range_center_at is not None:
            max_offset: float = max(
                abs(self._range_center_at - temp_min),
                abs(self._range_center_at - temp_max),
            )
            low_end = self._range_center_at - max_offset
            high_end = self._range_center_at + max_offset
        else:
            low_end = temp_min
            high_end = temp_max
        return [Palette.scale_lerp(scale, (low_end, high_end), value) for value in df[value_var]]

    @classmethod
    def get_name(cls) -> str:
        """Returns the name of this assigner strategy."""
        return "Assign by value"

    @classmethod
    def get_description(cls) -> str:
        """Returns the description of this assigner strategy."""
        return "Assigns colors based on the entry values, optionally within a specified range."

    def explain(self) -> str:
        hexes = self._hex_anchors()
        color_part = f"colors from {hexes[0]} to {hexes[-1]}"
        parts = [f"colorize them by their value in {color_part}"]
        if self._range_min is not None and self._range_max is not None:
            parts.append(f"restricted to values {self._range_min}\u2013{self._range_max}")
        elif self._range_min is not None:
            parts.append(f"from {self._range_min}")
        elif self._range_max is not None:
            parts.append(f"up to {self._range_max}")
        if self._range_center_at is not None:
            parts.append(f"centered at {self._range_center_at}")
        return " (".join(parts) + (")" if len(parts) > 1 else "")


class ColorRule:
    """
    Encapsulates a color rule for chart elements.
    """

    def __init__(
        self,
        assigner: ColorAssignerStrategy,
        selector: ColorSelectorStrategy | None = None,
    ) -> None:
        self._assigner = assigner
        self._selector = selector if selector is not None else ColorSelectorStrategy()

    @property
    def selector(self) -> ColorSelectorStrategy:
        """
        Gets the color selector strategy.
        """
        return self._selector

    @property
    def assigner(self) -> ColorAssignerStrategy:
        """
        Gets the color assigner strategy.
        """
        return self._assigner

    def explain(self) -> str:
        """
        Returns a human-readable description of what this color rule does.

        Combines the selector and assigner explanations into a single sentence,
        e.g. ``"selecting entries with a value above 12, colorize them by their
        value in colors from #00a000 to #00c000"``.

        :return: A prose description of the rule.
        :rtype: str
        """
        return f"{self._selector.explain()}, {self._assigner.explain()}"


@dataclass(frozen=True)
class ColorStrategyDescription:
    """
    Value object describing available color strategies.
    """

    supported_selectors: list[type[ColorSelectorStrategy]]
    supported_assigners: list[type[ColorAssignerStrategy]]


__all__ = [
    "ColorSelectorStrategy",
    "ColorSelectByCategoryName",
    "ColorSelectValuesBelow",
    "ColorSelectValuesAbove",
    "ColorSelectMaxValue",
    "ColorSelectMinValue",
    "ColorSelectValuesBelowMean",
    "ColorSelectValuesAboveMean",
    "ColorAssignerStrategy",
    "DiscreteColorAssignerStrategy",
    "ColorAssignerByCategory",
    "ColorAssignerByGroup",
    "ColorAssignerByValue",
    "ColorRule",
    "ColorStrategyDescription",
]



"""
Unit tests for ChoroplethMapChartStrategy._prepare_plot_data() and _render_from_payload().
"""

from typing import Any
from unittest.mock import MagicMock

import polars as pl
import pytest
from fields_metadata import FieldMetadata
from polychromos import HSLColor

from orama.color import (
    ColorAssignerByCategory,
    ColorRule,
    ColorSelectorStrategy,
)
from orama.strategies.choropleth_map import ChoroplethMapChartStrategy
from orama.variables import Aggregation


def _country_meta() -> MagicMock:
    m = MagicMock(spec=FieldMetadata)
    m.categorical = False
    m.numeric = False
    m.derived = False
    m.effective_type = str
    m.extra = {"suggested_validation": "iso_a2"}
    return m


def _make_choropleth(**kwargs: Any) -> ChoroplethMapChartStrategy:
    return ChoroplethMapChartStrategy(
        fields_metadata={"country": _country_meta()},
        country_variable="country",
        aggregation=Aggregation(name="count", function="count"),
        **kwargs,
    )


@pytest.fixture
def choropleth_df() -> pl.DataFrame:
    return pl.DataFrame({"country": ["US", "DE"], "count": [100, 200]})


def test_prepare_plot_data_returns_country_codes_names_values(
    choropleth_df: pl.DataFrame,
    tr: Any,
    theme: Any,
) -> None:
    instance = _make_choropleth()
    result = instance._prepare_plot_data({"main": choropleth_df}, tr, theme)
    # sorted alphabetically by country code
    assert result["country_codes"] == ["DE", "US"]
    assert len(result["country_names"]) == 2
    assert "Germany" in result["country_names"]
    assert "United States" in result["country_names"]
    assert len(result["values"]) == 2


def test_prepare_plot_data_continuous_mode_sets_color_scale(
    choropleth_df: pl.DataFrame,
    tr: Any,
    theme: Any,
) -> None:
    instance = _make_choropleth()
    result = instance._prepare_plot_data({"main": choropleth_df}, tr, theme)
    assert result["_use_discrete"] is False
    assert result["_color_scale"] is not None
    assert result["_discrete_colors"] is None


def test_prepare_plot_data_discrete_mode_with_categorical_rule(
    choropleth_df: pl.DataFrame,
    anchor_colors: list[HSLColor],
    tr: Any,
    theme: Any,
) -> None:
    rule = ColorRule(
        assigner=ColorAssignerByCategory(anchor_colors=anchor_colors),
        selector=ColorSelectorStrategy(),
    )
    instance = _make_choropleth(color_rules=[rule])
    result = instance._prepare_plot_data({"main": choropleth_df}, tr, theme)
    assert result["_use_discrete"] is True
    assert isinstance(result["_discrete_colors"], list)
    assert len(result["_discrete_colors"]) == 2
    assert all(c.startswith("#") for c in result["_discrete_colors"])


def _minimal_payload() -> dict[str, Any]:
    return {
        "country_names": ["Germany"],
        "values": [100],
        "_alpha3_codes": ["DEU"],
        "_use_discrete": True,
        "_discrete_colors": ["#ff0000"],
        "_color_scale": None,
        "title": "Test",
    }


@pytest.mark.parametrize("resolution_high,expected", [(True, 50), (False, 110)])
def test_render_resolution(resolution_high: bool, expected: int) -> None:
    instance = _make_choropleth(resolution_high=resolution_high)
    fig = instance._render_from_payload(_minimal_payload())
    assert fig.layout.geo.resolution == expected


@pytest.mark.parametrize("show_countries", [True, False])
def test_render_show_countries(show_countries: bool) -> None:
    instance = _make_choropleth(show_countries=show_countries)
    fig = instance._render_from_payload(_minimal_payload())
    assert fig.layout.geo.showcountries is show_countries
    assert fig.layout.geo.showlakes is show_countries


@pytest.mark.parametrize("show_ocean", [True, False])
def test_render_show_ocean(show_ocean: bool) -> None:
    instance = _make_choropleth(show_ocean=show_ocean)
    fig = instance._render_from_payload(_minimal_payload())
    assert fig.layout.geo.showocean is show_ocean



"""
Heat map figure strategy implementation.
"""

from typing import Any

import plotly.graph_objects as go
import polars as pl
from babel.support import Translations
from fields_metadata import FieldMetadata
from polychromos.palette import HSLColorScale, HSLColorSequence, Palette
from therismos.grouping import Aggregation as TherismosAggregation
from therismos.grouping import AggregationFunction

from orama.color import (
    ColorAssignerByValue,
    ColorAssignerStrategy,
    ColorRule,
    ColorSelectorStrategy,
)
from orama.options import (
    BooleanOption,
    ChartOptionsMap,
)
from orama.query import QueryParams
from orama.strategies.base import FigureStrategy
from orama.theme import Theme
from orama.variables import (
    Aggregation,
    CategoricalVariable,
    ChartVariablesMap,
    NumericalAggregationVariable,
)


class HeatMapChartStrategy(FigureStrategy):
    """
    Heat map figure strategy.
    """

    _variables: ChartVariablesMap = {
        "x_category": CategoricalVariable(
            name="X Category",
            description=(
                "Categorical variable(s) for the horizontal axis of the heat map. "
                "Each category corresponds to a heat map column. "
                "If several variables are provided, they will be concatenated in a cartesian "
                "product with all of their values to form the final categories."
            ),
            map_to="x_category_variables",
            multicategorical=True,
        ),
        "y_category": CategoricalVariable(
            name="Y Category",
            description=(
                "Categorical variable(s) for the vertical axis of the heat map. "
                "Each category corresponds to a heat map row. "
                "If several variables are provided, they will be concatenated in a cartesian "
                "product with all of their values to form the final categories."
            ),
            map_to="y_category_variables",
            multicategorical=True,
        ),
        "value": NumericalAggregationVariable(
            name="Value",
            description=("An aggregation of a numerical variable to produce the final value of each heat map cell."),
            map_to="aggregation",
        ),
    }
    _options: ChartOptionsMap = {
        "display_values": BooleanOption(
            name="Display values",
            description="Whether to display the numerical values in each cell of the heat map.",
            default_value=False,
            map_to="display_values",
        ),
    }

    def __init__(
        self,
        fields_metadata: dict[str, FieldMetadata],
        x_category_variables: list[str],
        y_category_variables: list[str],
        aggregation: Aggregation,
        display_values: bool = False,
        color_rules: list[ColorRule] | None = None,
        title: str | None = None,
        subtitle: str | None = None,
        caption: str | None = None,
        description: str | None = None,
        storytelling_context: str | None = None,
        x_label: str | None = None,
        y_label: str | None = None,
        value_label: str | None = None,
    ) -> None:
        super().__init__(
            fields_metadata=fields_metadata,
            color_rules=color_rules,
            title=title,
            subtitle=subtitle,
            caption=caption,
            description=description,
            storytelling_context=storytelling_context,
        )

        self._variables["x_category"].validate_fields(fields_metadata, *x_category_variables)
        self._variables["y_category"].validate_fields(fields_metadata, *y_category_variables)
        if aggregation.field:
            self._variables["value"].validate_fields(fields_metadata, aggregation.field)

        self._options["display_values"].validate(display_values)

        self._x_category_variables = x_category_variables
        self._y_category_variables = y_category_variables
        self._aggregation = aggregation
        self._display_values = display_values
        self._x_label = x_label
        self._y_label = y_label
        self._value_label = value_label

    @classmethod
    def name(cls) -> str:
        return "Heat Map"

    @classmethod
    def description(cls) -> str:
        return (
            "A heat map represents categorical data in a two-dimensional matrix. A color scale is "
            "used to represent the values of each cell in the matrix."
        )

    def get_query_params(self) -> dict[str, QueryParams]:
        return {
            "main": QueryParams(
                group_fields=[
                    *self._x_category_variables,
                    *self._y_category_variables,
                ],
                aggregations=[
                    TherismosAggregation(
                        id=self._aggregation.name,
                        function=AggregationFunction(self._aggregation.function),
                        field=self._aggregation.field,
                    ),
                ],
                dataframe_schema=pl.Schema(
                    self._get_schema_fields(),
                ),
            )
        }

    def _get_schema_fields(self) -> dict[str, type[pl.DataType]]:
        return {
            **{var: self._infer_column_type(self._fields_metadata, var) for var in self._x_category_variables},
            **{var: self._infer_column_type(self._fields_metadata, var) for var in self._y_category_variables},
            self._aggregation.name: (pl.Int64 if self._aggregation.function == AggregationFunction.COUNT else pl.Float64),
        }

    def _prepare_plot_data(
        self,
        dfs: dict[str, pl.DataFrame],
        tr: Translations,
        theme: Theme,
    ) -> dict[str, Any]:
        df: pl.DataFrame = dfs["main"]
        for category_variable in reversed(self._x_category_variables):
            df = self._sort_and_translate(df, category_variable, self._fields_metadata[category_variable], tr, sort_asc=True)
        df, x_alias = self._concatenate_variables(df, self._x_category_variables)
        for category_variable in reversed(self._y_category_variables):
            df = self._sort_and_translate(df, category_variable, self._fields_metadata[category_variable], tr, sort_asc=True)
        df, y_alias = self._concatenate_variables(df, self._y_category_variables)
        color_sequence: HSLColorSequence
        if self._color_rules:
            color_sequence = self._color_rules[0].assigner.anchor_colors
        else:
            color_sequence = theme.get_default_color_sequence()
        color_scale: HSLColorScale = Palette.to_color_scale(color_sequence)
        return {
            "x_categories": df[x_alias].to_list(),
            "y_categories": df[y_alias].to_list(),
            "values": df[self._aggregation.name].to_list(),
            "_color_scale": color_scale,
            "_x_alias": x_alias,
            "_y_alias": y_alias,
        }

    def _render_from_payload(self, payload: dict[str, Any]) -> go.Figure:
        color_scale = payload["_color_scale"]
        effective_title = payload["title"] or (
            f"Heat map of {self._aggregation.name} "
            f"by {', '.join(self._x_category_variables)} "
            f"and {', '.join(self._y_category_variables)}"
        )
        effective_x = self._x_label or ", ".join(self._x_category_variables)
        effective_y = self._y_label or ", ".join(self._y_category_variables)
        effective_value = self._value_label or self._append_unit(self._aggregation.default_label, self._aggregation.field)
        fig = go.Figure(
            data=go.Heatmap(
                z=payload["values"],
                x=payload["x_categories"],
                y=payload["y_categories"],
                colorscale=Palette.color_scale_to_plotly(color_scale),
                texttemplate="%{z:.2f}" if self._display_values else None,
            )
        )
        fig.update_traces(colorbar_title_text=effective_value)
        fig.update_layout(xaxis_title=effective_x, yaxis_title=effective_y)
        for color_rule in self._color_rules:
            if isinstance(color_rule.assigner, ColorAssignerByValue) and color_rule.assigner.range_min is not None:
                fig.update_traces(zmin=color_rule.assigner.range_min)
                break
        for color_rule in self._color_rules:
            if isinstance(color_rule.assigner, ColorAssignerByValue) and color_rule.assigner.range_max is not None:
                fig.update_traces(zmax=color_rule.assigner.range_max)
                break
        for color_rule in self._color_rules:
            if isinstance(color_rule.assigner, ColorAssignerByValue) and color_rule.assigner.range_center_at is not None:
                fig.update_traces(zmid=color_rule.assigner.range_center_at)
                break
        self._apply_title(fig, effective_title, payload.get("subtitle"))
        return fig

    @classmethod
    def supported_color_selectors(cls) -> list[type[ColorSelectorStrategy]]:
        return [
            ColorSelectorStrategy,
        ]

    @classmethod
    def supported_color_assigners(cls) -> list[type[ColorAssignerStrategy]]:
        return [
            ColorAssignerByValue,
        ]


__all__ = [
    "HeatMapChartStrategy",
]



"""
Bar chart figure strategy implementation.
"""

import copy
from typing import Any

import plotly.express as px
import plotly.graph_objects as go
import polars as pl
from babel.support import Translations
from fields_metadata import FieldMetadata
from polychromos.palette import HSLColorSequence
from therismos.grouping import Aggregation as TherismosAggregation
from therismos.grouping import AggregationFunction

from orama.color import (
    ColorAssignerByCategory,
    ColorAssignerByValue,
    ColorAssignerStrategy,
    ColorRule,
    ColorSelectByCategoryName,
    ColorSelectMaxValue,
    ColorSelectMinValue,
    ColorSelectorStrategy,
    ColorSelectValuesAbove,
    ColorSelectValuesAboveMean,
    ColorSelectValuesBelow,
    ColorSelectValuesBelowMean,
)
from orama.enums import SortOrder
from orama.options import (
    BooleanOption,
    ChartOptionsMap,
    FloatOption,
    IntOption,
    SortOrderOption,
    StringOption,
)
from orama.query import QueryParams
from orama.strategies.base import FigureStrategy
from orama.theme import Theme
from orama.variables import (
    Aggregation,
    CategoricalVariable,
    ChartVariablesMap,
    NumericalAggregationVariable,
)

_OTHERS_SIMPLE_FUNCTIONS = {
    AggregationFunction.COUNT,
    AggregationFunction.SUM,
    AggregationFunction.MIN,
    AggregationFunction.MAX,
}


class AbstractBarChartStrategy(FigureStrategy):
    """
    Abstract base class for bar chart figure strategies.
    """

    def __init__(
        self,
        fields_metadata: dict[str, FieldMetadata],
        category_variables: list[str],
        aggregation: Aggregation,
        color_rules: list[ColorRule] | None = None,
        title: str | None = None,
        subtitle: str | None = None,
        caption: str | None = None,
        description: str | None = None,
        storytelling_context: str | None = None,
        category_label: str | None = None,
        value_label: str | None = None,
    ) -> None:
        super().__init__(
            fields_metadata=fields_metadata,
            color_rules=color_rules,
            title=title,
            subtitle=subtitle,
            caption=caption,
            description=description,
            storytelling_context=storytelling_context,
        )

        self._variables["category"].validate_fields(fields_metadata, *category_variables)
        if aggregation.field:
            self._variables["value"].validate_fields(fields_metadata, aggregation.field)

        self._category_variables = category_variables
        self._aggregation = aggregation
        self._category_label = category_label
        self._value_label = value_label

    def get_query_params(self) -> dict[str, QueryParams]:
        return {
            "main": QueryParams(
                group_fields=copy.copy(self._category_variables),
                aggregations=[
                    TherismosAggregation(
                        id=self._aggregation.name,
                        function=AggregationFunction(self._aggregation.function),
                        field=self._aggregation.field,
                    ),
                ],
                dataframe_schema=pl.Schema(
                    self._get_schema_fields(),
                ),
            )
        }

    def _get_schema_fields(self) -> dict[str, type[pl.DataType]]:
        return {
            **{var: self._infer_column_type(self._fields_metadata, var) for var in self._category_variables},
            self._aggregation.name: (pl.Int64 if self._aggregation.function == AggregationFunction.COUNT else pl.Float64),
        }


class BarChartStrategy(AbstractBarChartStrategy):
    """
    Bar chart figure strategy.
    """

    _variables: ChartVariablesMap = {
        "category": CategoricalVariable(
            name="Category",
            description=(
                "Categorical variable(s) for the bar chart. "
                "Each category corresponds to a bar. "
                "If several variables are provided, they will be concatenated in a cartesian "
                "product with all of their values to form the final categories."
            ),
            map_to="category_variables",
            multicategorical=True,
        ),
        "value": NumericalAggregationVariable(
            name="Value",
            description=(
                "An aggregation of a numerical variable to produce the final value of each "
                "category. This value is proportional to the length of each bar."
            ),
            map_to="aggregation",
        ),
    }
    _options: ChartOptionsMap = {
        "sort_by_value": SortOrderOption(
            name="Sort by value",
            description=(
                "Sorting order of the bars according to their value. If none, bars are sorted according to the category variable(s)."
            ),
            map_to="sort_by_value",
            default_value=SortOrder.NONE,
        ),
        "horizontal": BooleanOption(
            name="Horizontal bars",
            description="When enabled, bars are rendered horizontally instead of vertically.",
            map_to="horizontal",
            default_value=False,
        ),
        "min_value": FloatOption(
            name="Minimum value",
            description="Minimum value for the bar chart value axis.",
            map_to="min_value",
            default_value=None,
        ),
        "max_value": FloatOption(
            name="Maximum value",
            description="Maximum value for the bar chart value axis.",
            map_to="max_value",
            default_value=None,
        ),
        "mean_category_name": StringOption(
            name="Mean category name",
            description=(
                "The name to use for the category that represents the mean of every category. "
                "If not provided, the mean category is not shown. When the value aggregation "
                "function is the count, the mean category represents the average count per "
                "category. Otherwise, it represents the mean value weighted by the count of each "
                "category."
            ),
            map_to="mean_category_name",
            default_value="",
        ),
        "mean_annotation": BooleanOption(
            name="Mean annotation",
            description=(
                "When enabled, shows an annotation with the mean value on the bar chart. "
                "Only applicable if a mean category name is provided. The mean value is "
                "calculated as the overall average when the value aggregation function is the "
                "count; otherwise, it is the mean value weighted by the count of each category."
            ),
            map_to="mean_annotation",
            default_value=False,
        ),
        "others_threshold": FloatOption(
            name="Others threshold",
            description=(
                "When set, all bars whose value is strictly below this threshold are grouped into a "
                "single bar. Only applicable when sorting is active. "
                "Mutually exclusive with 'others_top_n'."
            ),
            map_to="others_threshold",
            default_value=None,
        ),
        "others_top_n": IntOption(
            name="Others top N",
            description=(
                "When set, only the top N bars by value are shown individually; all remaining bars "
                "are grouped into a single bar. Only applicable when sorting is active. "
                "Mutually exclusive with 'others_threshold'."
            ),
            map_to="others_top_n",
            min_value=1,
            default_value=None,
        ),
        "others_category_name": StringOption(
            name="Others category name",
            description="Label for the grouped bar. Defaults to 'Others'.",
            map_to="others_category_name",
            default_value="Others",
        ),
    }

    def __init__(
        self,
        fields_metadata: dict[str, FieldMetadata],
        category_variables: list[str],
        aggregation: Aggregation,
        sort_by_value: SortOrder = SortOrder.NONE,
        horizontal: bool = False,
        min_value: float | None = None,
        max_value: float | None = None,
        mean_category_name: str = "",
        mean_annotation: bool = False,
        others_threshold: float | None = None,
        others_top_n: int | None = None,
        others_category_name: str = "Others",
        color_rules: list[ColorRule] | None = None,
        title: str | None = None,
        subtitle: str | None = None,
        caption: str | None = None,
        description: str | None = None,
        storytelling_context: str | None = None,
        category_label: str | None = None,
        value_label: str | None = None,
    ) -> None:
        super().__init__(
            fields_metadata=fields_metadata,
            category_variables=category_variables,
            aggregation=aggregation,
            color_rules=color_rules,
            title=title,
            subtitle=subtitle,
            caption=caption,
            description=description,
            storytelling_context=storytelling_context,
            category_label=category_label,
            value_label=value_label,
        )

        self._options["sort_by_value"].validate(sort_by_value)
        self._options["horizontal"].validate(horizontal)
        if min_value is not None:
            self._options["min_value"].validate(min_value)
        if max_value is not None:
            self._options["max_value"].validate(max_value)
        if mean_category_name is not None:
            self._options["mean_category_name"].validate(mean_category_name)
        self._options["mean_annotation"].validate(mean_annotation)
        if others_threshold is not None and others_top_n is not None:
            raise ValueError("'others_threshold' and 'others_top_n' are mutually exclusive")
        if others_threshold is not None and sort_by_value == SortOrder.NONE:
            raise ValueError("'others_threshold' requires sorting to be active (sort_by_value != SortOrder.NONE)")
        if others_top_n is not None and sort_by_value == SortOrder.NONE:
            raise ValueError("'others_top_n' requires sorting to be active (sort_by_value != SortOrder.NONE)")
        if others_category_name and mean_category_name and others_category_name == mean_category_name:
            raise ValueError(f"'others_category_name' and 'mean_category_name' must not be the same: '{others_category_name}'")
        if others_threshold is not None:
            self._options["others_threshold"].validate(others_threshold)
        if others_top_n is not None:
            self._options["others_top_n"].validate(others_top_n)
        self._options["others_category_name"].validate(others_category_name)

        self._sort_by_value = sort_by_value
        self._horizontal = horizontal
        self._min_value = min_value
        self._max_value = max_value
        self._mean_category_name = mean_category_name
        self._mean_annotation = mean_annotation
        self._others_threshold = others_threshold
        self._others_top_n = others_top_n
        self._others_category_name = others_category_name

    @classmethod
    def name(cls) -> str:
        return "Bar Chart"

    @classmethod
    def description(cls) -> str:
        return (
            "A bar chart represents categorical data with rectangular bars with heights or "
            "lengths proportional to the values that they represent. The bars can be plotted "
            "vertically or horizontally."
        )

    @property
    def _requires_mean(self) -> bool:
        return any(
            [
                self._mean_category_name,
                self._mean_annotation,
                *[
                    color_rule.selector.__class__
                    in {
                        ColorSelectValuesBelowMean,
                        ColorSelectValuesAboveMean,
                    }
                    for color_rule in (self._color_rules or [])
                ],
            ]
        )

    @property
    def _requires_total_count(self) -> bool:
        others_need_count = (
            self._others_threshold is not None or self._others_top_n is not None
        ) and self._aggregation.function not in _OTHERS_SIMPLE_FUNCTIONS
        return self._requires_mean or others_need_count

    def get_query_params(self) -> dict[str, QueryParams]:
        query_params = super().get_query_params()
        if self._requires_total_count:
            query_params["main"].aggregations.append(
                TherismosAggregation(
                    id="__total_count__",
                    function=AggregationFunction.COUNT,
                ),
            )
        return query_params

    def _get_schema_fields(self) -> dict[str, type[pl.DataType]]:
        fields = super()._get_schema_fields()
        if self._requires_total_count:
            fields["__total_count__"] = pl.Int64
        return fields

    def _apply_others_grouping(self, df: pl.DataFrame, alias: str) -> pl.DataFrame:
        """
        Collapses low-signal bars into a single "Others" bar.

        :param df: The DataFrame to process.
        :type df: pl.DataFrame
        :param alias: The column name for the category axis.
        :type alias: str
        :return: The processed DataFrame with an "Others" bar appended if applicable.
        :rtype: pl.DataFrame
        """
        agg_name = self._aggregation.name
        if self._mean_category_name:
            mean_rows = df.filter(pl.col(alias) == self._mean_category_name)
            data_rows = df.filter(pl.col(alias) != self._mean_category_name)
        else:
            mean_rows = df.head(0)
            data_rows = df

        if self._others_threshold is not None:
            others_rows = data_rows.filter(pl.col(agg_name) < self._others_threshold)
            kept_rows = data_rows.filter(pl.col(agg_name) >= self._others_threshold)
        else:
            assert self._others_top_n is not None
            sorted_desc = data_rows.sort(agg_name, descending=True)
            kept_rows_unordered = sorted_desc.head(self._others_top_n)
            others_rows = sorted_desc.slice(self._others_top_n)
            kept_indices = kept_rows_unordered[alias].to_list()
            kept_rows = data_rows.filter(pl.col(alias).is_in(kept_indices))

        if others_rows.is_empty():
            return pl.concat([data_rows, mean_rows])

        fn = self._aggregation.function
        if fn in (AggregationFunction.COUNT, AggregationFunction.SUM):
            others_value = others_rows[agg_name].sum()
        elif fn == AggregationFunction.MIN:
            others_value = others_rows[agg_name].min()  # type: ignore[assignment]
        elif fn == AggregationFunction.MAX:
            others_value = others_rows[agg_name].max()  # type: ignore[assignment]
        else:
            weighted_sum = (others_rows[agg_name] * others_rows["__total_count__"]).sum()
            total_weight = others_rows["__total_count__"].sum()
            others_value = weighted_sum / total_weight  # type: ignore[operator]

        others_row_data: dict[str, list[Any]] = {
            alias: [self._others_category_name],
            agg_name: [others_value],
        }
        if "__total_count__" in df.columns:
            others_row_data["__total_count__"] = [others_rows["__total_count__"].sum()]
        if "__total_mean__" in df.columns:
            others_row_data["__total_mean__"] = [df["__total_mean__"][0]]

        others_row = pl.DataFrame(others_row_data).select(df.columns)
        return pl.concat([kept_rows, others_row, mean_rows], how="vertical_relaxed")

    def _prepare_df(
        self,
        df: pl.DataFrame,
        tr: Translations,
    ) -> tuple[pl.DataFrame, str, float | None]:
        """
        Sort, translate, compute the mean row (if needed), sort by value, and apply others grouping.

        :param df: Raw DataFrame from the query result.
        :type df: pl.DataFrame
        :param tr: Translations for localizing category labels.
        :type tr: Translations
        :return: Processed DataFrame, category alias column name, and optional mean value.
        :rtype: tuple[pl.DataFrame, str, float | None]
        """
        for category_variable in reversed(self._category_variables):
            df = self._sort_and_translate(
                df,
                category_variable,
                self._fields_metadata[category_variable],
                tr,
                sort_asc=True,
            )
        df, alias = self._concatenate_variables(df, self._category_variables)
        mean_value: float | None = None
        if self._requires_mean:
            df = df.with_columns(pl.col(self._aggregation.name).cast(pl.Float64))
            total_count_value = df["__total_count__"].sum()
            if self._aggregation.function == AggregationFunction.COUNT:
                mean_value = float(df[self._aggregation.name].mean())  # type: ignore[arg-type]
            else:
                weighted_sum = float((df[self._aggregation.name] * df["__total_count__"]).sum())
                mean_value = float(weighted_sum / total_count_value)  # type: ignore[operator]
            if self._mean_category_name:
                df = pl.concat(
                    [
                        df,
                        pl.DataFrame(
                            {
                                alias: [self._mean_category_name],
                                self._aggregation.name: [mean_value],
                                "__total_count__": [total_count_value],
                            }
                        ).select(df.columns),
                    ],
                    how="vertical_relaxed",
                )
            df = df.with_columns(pl.lit(mean_value).alias("__total_mean__"))
        if self._sort_by_value != SortOrder.NONE:
            df = df.sort(
                self._aggregation.name,
                descending=self._sort_by_value == SortOrder.DESCENDING,
            )
        if self._others_threshold is not None or self._others_top_n is not None:
            df = self._apply_others_grouping(df, alias)
        return df, alias, mean_value

    def _compute_bar_colors(
        self,
        df: pl.DataFrame,
        alias: str,
        theme: Theme,
    ) -> HSLColorSequence:
        """
        Produce the per-bar color sequence via color rules or the default category assigner.

        :param df: Prepared DataFrame (post :meth:`_prepare_df`).
        :type df: pl.DataFrame
        :param alias: Category axis column name.
        :type alias: str
        :param theme: Theme providing color sequences.
        :type theme: Theme
        :return: A color sequence with one entry per bar.
        :rtype: HSLColorSequence
        """
        if self._color_rules:
            return self._apply_color_rules(df, alias, self._aggregation.name, theme)
        return ColorAssignerByCategory(
            anchor_colors=theme.get_default_color_sequence(),
        ).assign_colors(df, alias, self._aggregation.name)

    def _configure_bar_layout(
        self,
        fig: go.Figure,
        mean_value: float | None,
        effective_category: str,
        effective_value: str,
        annotation_color: str | None,
    ) -> None:
        """
        Apply axis titles, value range, legend visibility, and optional mean annotation.

        :param fig: Plotly figure to update.
        :type fig: go.Figure
        :param mean_value: Overall mean used for the annotation line, or ``None``.
        :type mean_value: float | None
        :param effective_category: Resolved category axis label.
        :type effective_category: str
        :param effective_value: Resolved value axis label.
        :type effective_value: str
        :param annotation_color: Hex color for the mean annotation line, or ``None``.
        :type annotation_color: str | None
        """
        fig.update_layout(showlegend=False)
        value_range = (self._min_value, self._max_value)
        if self._horizontal:
            fig.update_layout(xaxis_title=effective_value, yaxis_title=effective_category)
            fig.update_xaxes(range=value_range)
            if self._mean_annotation and mean_value is not None:
                fig.add_vline(
                    x=mean_value,
                    line_color=annotation_color,
                    line_width=3,
                    opacity=0.7,
                    layer="below",
                )
        else:
            fig.update_layout(xaxis_title=effective_category, yaxis_title=effective_value)
            fig.update_yaxes(range=value_range)
            if self._mean_annotation and mean_value is not None:
                fig.add_hline(
                    y=mean_value,
                    line_color=annotation_color,
                    line_width=3,
                    opacity=0.7,
                    layer="below",
                )

    def _prepare_plot_data(
        self,
        dfs: dict[str, pl.DataFrame],
        tr: Translations,
        theme: Theme,
    ) -> dict[str, Any]:
        df, alias, mean_value = self._prepare_df(dfs["main"], tr)
        bar_colors = self._compute_bar_colors(df, alias, theme)
        return {
            "categories": df[alias].to_list(),
            "values": df[self._aggregation.name].to_list(),
            "colors": [c.to_css_hex() for c in bar_colors],
            "mean_value": mean_value,
            "_alias": alias,
            "_annotation_color": theme.get_color("secondary_color_softer").to_css_hex(),
        }

    def _render_from_payload(self, payload: dict[str, Any]) -> go.Figure:
        alias = payload["_alias"]
        categories = payload["categories"]
        values = payload["values"]
        colors = payload["colors"]
        mean_value = payload["mean_value"]
        df = pl.DataFrame({alias: categories, self._aggregation.name: values})
        x_var = alias
        y_var = self._aggregation.name
        if self._horizontal:
            x_var, y_var = y_var, x_var
        color_map = dict(zip(categories, colors))
        fig = px.bar(
            df.to_pandas(),
            x=x_var,
            y=y_var,
            color=alias,
            color_discrete_map=color_map,
            orientation="h" if self._horizontal else "v",
        )
        effective_title = payload["title"] or (f"Bar chart of {self._aggregation.name} by {', '.join(self._category_variables)}")
        effective_category = self._category_label or ", ".join(self._category_variables)
        effective_value = self._value_label or self._append_unit(self._aggregation.default_label, self._aggregation.field)
        self._configure_bar_layout(fig, mean_value, effective_category, effective_value, payload["_annotation_color"])
        self._apply_title(fig, effective_title, payload.get("subtitle"))
        return fig

    @classmethod
    def supported_color_selectors(cls) -> list[type[ColorSelectorStrategy]]:
        return [
            ColorSelectorStrategy,
            ColorSelectByCategoryName,
            ColorSelectValuesBelow,
            ColorSelectValuesAbove,
            ColorSelectMaxValue,
            ColorSelectMinValue,
            ColorSelectValuesBelowMean,
            ColorSelectValuesAboveMean,
        ]

    @classmethod
    def supported_color_assigners(cls) -> list[type[ColorAssignerStrategy]]:
        return [
            ColorAssignerByCategory,
            ColorAssignerByValue,
        ]


__all__ = [
    "AbstractBarChartStrategy",
    "BarChartStrategy",
]



"""
Grouped boxplot chart figure strategy implementation.
"""

from typing import Any

import plotly.graph_objects as go
import polars as pl
from babel.support import Translations
from fields_metadata import FieldMetadata
from polychromos.palette import HSLColorSequence, Palette

from orama.color import (
    ColorAssignerByGroup,
    ColorAssignerStrategy,
    ColorRule,
    ColorSelectByCategoryName,
    ColorSelectorStrategy,
)
from orama.options import (
    BooleanOption,
    ChartOptionsMap,
)
from orama.query import QueryParams
from orama.strategies.boxplot import AbstractBoxplotChartStrategy
from orama.theme import Theme
from orama.variables import (
    CategoricalVariable,
    ChartVariablesMap,
    NonDerivedNumericalVariable,
)


class GroupedBoxplotChartStrategy(AbstractBoxplotChartStrategy):
    """
    Grouped boxplot chart figure strategy.

    Extends the standard boxplot chart with an additional grouping dimension, producing
    side-by-side boxplots per category — one per unique value of the grouping variable.
    """

    _variables: ChartVariablesMap = {
        "category": CategoricalVariable(
            name="Category",
            description=(
                "Categorical variable(s) for the grouped boxplot chart. "
                "Each category corresponds to a set of boxplots. "
                "If several variables are provided, they will be concatenated in a cartesian "
                "product with all of their values to form the final categories."
            ),
            map_to="category_variables",
            multicategorical=True,
        ),
        "value": NonDerivedNumericalVariable(
            name="Value",
            description="A numerical variable to extract the boxplot statistics from for each category.",
            map_to="value_variable",
        ),
        "grouping": CategoricalVariable(
            name="Grouping",
            description=(
                "Categorical variable for grouping the boxplots. "
                "Categories are split into several groups, one per each value of this variable."
            ),
            map_to="group_variable",
            multicategorical=False,
        ),
    }
    _options: ChartOptionsMap = {
        "horizontal": BooleanOption(
            name="Horizontal boxplots",
            description="When enabled, boxplots are rendered horizontally instead of vertically.",
            map_to="horizontal",
            default_value=False,
        ),
        "display_mean": BooleanOption(
            name="Display mean",
            description="Whether to display the mean value in each boxplot.",
            map_to="display_mean",
            default_value=False,
        ),
        "display_sd": BooleanOption(
            name="Display standard deviation",
            description=("Whether to display the standard deviation in each boxplot. Only applies if mean is also displayed."),
            map_to="display_sd",
            default_value=False,
        ),
    }

    def __init__(
        self,
        fields_metadata: dict[str, FieldMetadata],
        category_variables: list[str],
        value_variable: str,
        group_variable: str,
        horizontal: bool = False,
        display_mean: bool = False,
        display_sd: bool = False,
        color_rules: list[ColorRule] | None = None,
        title: str | None = None,
        subtitle: str | None = None,
        caption: str | None = None,
        description: str | None = None,
        storytelling_context: str | None = None,
        category_label: str | None = None,
        value_label: str | None = None,
        group_label: str | None = None,
    ) -> None:
        super().__init__(
            fields_metadata=fields_metadata,
            category_variables=category_variables,
            value_variable=value_variable,
            color_rules=color_rules,
            title=title,
            subtitle=subtitle,
            caption=caption,
            description=description,
            storytelling_context=storytelling_context,
            category_label=category_label,
            value_label=value_label,
        )

        self._variables["grouping"].validate_fields(fields_metadata, group_variable)

        self._options["horizontal"].validate(horizontal)
        self._options["display_mean"].validate(display_mean)
        self._options["display_sd"].validate(display_sd)

        self._group_variable = group_variable
        self._horizontal = horizontal
        self._display_mean = display_mean
        self._display_sd = display_sd & self._display_mean
        self._group_label = group_label

    @classmethod
    def name(cls) -> str:
        """Return the human-readable name of this strategy."""
        return "Grouped Boxplot Chart"

    @classmethod
    def description(cls) -> str:
        """Return a description of this strategy."""
        return (
            "A grouped boxplot chart represents statistical distribution of samples segregated by "
            "categories, with an additional grouping dimension. For each category, multiple "
            "side-by-side boxplots are displayed — one per unique value of the grouping variable. "
            "Each boxplot shows the first quartile (Q1), median (Q2), and third quartile (Q3) of "
            "the samples, as well as the computed fences (whiskers) at 1.5 times the "
            "interquartile range (IQR) from the mean."
        )

    def get_query_params(self) -> dict[str, QueryParams]:
        """
        Return the query parameters for this strategy.

        :return: Dictionary mapping query names to their parameters.
        :rtype: dict[str, QueryParams]
        """
        params = super().get_query_params()
        params["main"].group_fields.append(self._group_variable)
        return params

    def _get_schema_fields(self) -> dict[str, type[pl.DataType]]:
        fields = super()._get_schema_fields()
        fields[self._group_variable] = self._infer_column_type(self._fields_metadata, self._group_variable)
        return fields

    def _prepare_plot_data(
        self,
        dfs: dict[str, pl.DataFrame],
        tr: Translations,
        theme: Theme,
    ) -> dict[str, Any]:
        df: pl.DataFrame = self._compute_fences(dfs["main"])
        df = self._sort_and_translate(
            df,
            self._group_variable,
            self._fields_metadata[self._group_variable],
            tr,
            sort_asc=True,
        )
        for category_variable in reversed(self._category_variables):
            df = self._sort_and_translate(
                df,
                category_variable,
                self._fields_metadata[category_variable],
                tr,
                sort_asc=True,
            )
        df, alias = self._concatenate_variables(df, self._category_variables)

        sorted_groups = df[self._group_variable].unique(maintain_order=True).to_list()
        group_df = pl.DataFrame({self._group_variable: sorted_groups})
        n_groups = len(sorted_groups)

        group_colors: HSLColorSequence
        if self._color_rules:
            group_colors = [theme.get_color("primary_color_normal")] * n_groups
            for color_rule in self._color_rules:
                selector = color_rule.selector.selector(group_df, self._group_variable, "mean")
                assigned_colors = color_rule.assigner.assign_colors(group_df, self._group_variable, "mean", self._group_variable)
                group_colors = Palette.mix_color_sequences(
                    group_colors,
                    assigned_colors,
                    selector,
                    indexing=Palette.MixIndexing.BY_POSITION,
                )
        else:
            group_colors = ColorAssignerByGroup(
                anchor_colors=theme.get_default_color_sequence(),
            ).assign_colors(group_df, self._group_variable, "mean", self._group_variable)

        return {
            "categories": df[alias].unique(maintain_order=True).to_list(),
            "groups": sorted_groups,
            "color_map": {str(g): group_colors[i].to_css_hex() for i, g in enumerate(sorted_groups)},
            "_df": df,
            "_alias": alias,
            "_group_colors": group_colors,
        }

    def _render_from_payload(self, payload: dict[str, Any]) -> go.Figure:
        df = payload["_df"]
        alias = payload["_alias"]
        group_colors = payload["_group_colors"]
        sorted_groups = payload["groups"]
        data = []
        for i, group in enumerate(sorted_groups):
            group_data = df.filter(pl.col(self._group_variable) == group)
            box_kwargs: dict[str, Any] = dict(
                name=str(group),
                q1=group_data["q1"].to_list(),
                median=group_data["median"].to_list(),
                q3=group_data["q3"].to_list(),
                lowerfence=group_data["lowerfence"].to_list(),
                upperfence=group_data["upperfence"].to_list(),
                boxpoints="outliers",
                marker=dict(color=group_colors[i].to_css_hex()),
                showlegend=True,
            )
            if self._display_mean:
                box_kwargs["mean"] = group_data["mean"].to_list()
                if self._display_sd:
                    box_kwargs["sd"] = group_data["sd"].to_list()
            if self._horizontal:
                data.append(go.Box(y=group_data[alias].to_list(), orientation="h", **box_kwargs))
            else:
                data.append(go.Box(x=group_data[alias].to_list(), **box_kwargs))
        effective_title = payload["title"] or (
            f"Grouped boxplot of {self._value_variable} by {', '.join(self._category_variables)} and {self._group_variable}"
        )
        effective_category = self._category_label or ", ".join(self._category_variables)
        effective_value = self._value_label or self._append_unit(self._value_variable, self._value_variable)
        effective_group = self._group_label or self._group_variable
        fig = go.Figure(data=data)
        if self._horizontal:
            fig.update_layout(xaxis_title=effective_value, yaxis_title=effective_category)
        else:
            fig.update_layout(xaxis_title=effective_category, yaxis_title=effective_value)
        fig.update_layout(boxmode="group", legend_title_text=effective_group)
        self._apply_title(fig, effective_title, payload.get("subtitle"))
        return fig

    @classmethod
    def supported_color_selectors(cls) -> list[type[ColorSelectorStrategy]]:
        """Return the list of supported color selector strategies."""
        return [
            ColorSelectorStrategy,
            ColorSelectByCategoryName,
        ]

    @classmethod
    def supported_color_assigners(cls) -> list[type[ColorAssignerStrategy]]:
        """Return the list of supported color assigner strategies."""
        return [
            ColorAssignerByGroup,
        ]


__all__ = ["GroupedBoxplotChartStrategy"]



"""
Unit tests for HeatMapChartStrategy._prepare_plot_data().
"""

from typing import Any
from unittest.mock import MagicMock

import polars as pl
import pytest
from fields_metadata import FieldMetadata

from orama.strategies.heatmap import HeatMapChartStrategy
from orama.variables import Aggregation


def _cat_meta() -> MagicMock:
    m = MagicMock(spec=FieldMetadata)
    m.categorical = True
    m.numeric = False
    m.derived = False
    m.effective_type = str
    m.extra = {}
    return m


def _make_heatmap(**kwargs: Any) -> HeatMapChartStrategy:
    return HeatMapChartStrategy(
        fields_metadata={"x": _cat_meta(), "y": _cat_meta()},
        x_category_variables=["x"],
        y_category_variables=["y"],
        aggregation=Aggregation(name="count", function="count"),
        **kwargs,
    )


@pytest.fixture
def heatmap_df() -> pl.DataFrame:
    return pl.DataFrame(
        {
            "x": ["A", "A", "B", "B"],
            "y": ["P", "Q", "P", "Q"],
            "count": [1, 2, 3, 4],
        }
    )


@pytest.fixture
def two_color_theme(theme: Any, anchor_color: Any) -> Any:
    """Theme mock with two anchor colors so Palette.to_color_scale doesn't raise."""
    from polychromos import HSLColor

    second = HSLColor(hue=0.5, saturation=0.5, lightness=0.5)
    theme.get_default_color_sequence.return_value = [anchor_color, second]
    return theme


def test_prepare_plot_data_returns_x_y_categories_and_values(
    heatmap_df: pl.DataFrame,
    tr: Any,
    two_color_theme: Any,
) -> None:
    heatmap = _make_heatmap()
    result = heatmap._prepare_plot_data({"main": heatmap_df}, tr, two_color_theme)
    assert len(result["x_categories"]) == 4
    assert len(result["y_categories"]) == 4
    assert len(result["values"]) == 4
    assert set(result["values"]) == {1, 2, 3, 4}


def test_prepare_plot_data_color_scale_present(
    heatmap_df: pl.DataFrame,
    tr: Any,
    two_color_theme: Any,
) -> None:
    heatmap = _make_heatmap()
    result = heatmap._prepare_plot_data({"main": heatmap_df}, tr, two_color_theme)
    assert result["_color_scale"] is not None



"""
Boxplot chart figure strategy implementation.
"""

import copy
from typing import Any

import plotly.graph_objects as go
import polars as pl
from babel.support import Translations
from fields_metadata import FieldMetadata
from polychromos.palette import HSLColorSequence, Palette
from therismos.grouping import Aggregation as TherismosAggregation
from therismos.grouping import AggregationFunction

from orama.color import (
    ColorAssignerByCategory,
    ColorAssignerByValue,
    ColorAssignerStrategy,
    ColorRule,
    ColorSelectByCategoryName,
    ColorSelectMaxValue,
    ColorSelectMinValue,
    ColorSelectorStrategy,
    ColorSelectValuesAbove,
    ColorSelectValuesAboveMean,
    ColorSelectValuesBelow,
    ColorSelectValuesBelowMean,
)
from orama.options import (
    BooleanOption,
    ChartOptionsMap,
    StringOption,
)
from orama.query import QueryParams
from orama.strategies.base import FigureStrategy
from orama.theme import Theme
from orama.variables import (
    CategoricalVariable,
    ChartVariablesMap,
    NonDerivedNumericalVariable,
)


class AbstractBoxplotChartStrategy(FigureStrategy):
    """
    Abstract base class for boxplot chart figure strategies.
    """

    def __init__(
        self,
        fields_metadata: dict[str, FieldMetadata],
        category_variables: list[str],
        value_variable: str,
        color_rules: list[ColorRule] | None = None,
        title: str | None = None,
        subtitle: str | None = None,
        caption: str | None = None,
        description: str | None = None,
        storytelling_context: str | None = None,
        category_label: str | None = None,
        value_label: str | None = None,
    ) -> None:
        super().__init__(
            fields_metadata=fields_metadata,
            color_rules=color_rules,
            title=title,
            subtitle=subtitle,
            caption=caption,
            description=description,
            storytelling_context=storytelling_context,
        )

        self._variables["category"].validate_fields(fields_metadata, *category_variables)
        self._variables["value"].validate_fields(fields_metadata, value_variable)

        self._category_variables = category_variables
        self._value_variable = value_variable
        self._category_label = category_label
        self._value_label = value_label

    def get_query_params(self) -> dict[str, QueryParams]:
        return {
            "main": QueryParams(
                group_fields=copy.copy(self._category_variables),
                aggregations=[
                    TherismosAggregation(
                        id="q1",
                        function=AggregationFunction.Q1,
                        field=self._value_variable,
                    ),
                    TherismosAggregation(
                        id="median",
                        function=AggregationFunction.MEDIAN,
                        field=self._value_variable,
                    ),
                    TherismosAggregation(
                        id="q3",
                        function=AggregationFunction.Q3,
                        field=self._value_variable,
                    ),
                    TherismosAggregation(
                        id="mean",
                        function=AggregationFunction.AVERAGE,
                        field=self._value_variable,
                    ),
                    TherismosAggregation(
                        id="sd",
                        function=AggregationFunction.STDDEV,
                        field=self._value_variable,
                    ),
                    TherismosAggregation(
                        id="min",
                        function=AggregationFunction.MIN,
                        field=self._value_variable,
                    ),
                    TherismosAggregation(
                        id="max",
                        function=AggregationFunction.MAX,
                        field=self._value_variable,
                    ),
                ],
                dataframe_schema=pl.Schema(
                    self._get_schema_fields(),
                ),
            )
        }

    def _get_schema_fields(self) -> dict[str, type[pl.DataType]]:
        return {
            **{var: self._infer_column_type(self._fields_metadata, var) for var in self._category_variables},
            "q1": pl.Float64,
            "median": pl.Float64,
            "q3": pl.Float64,
            "mean": pl.Float64,
            "sd": pl.Float64,
            "min": pl.Float64,
            "max": pl.Float64,
        }

    def _compute_fences(
        self,
        df: pl.DataFrame,
    ) -> pl.DataFrame:
        iqr = df["q3"] - df["q1"]
        lower_fence: pl.Series = pl.Series(
            name="lowerfence",
            values=[max(df["min"][i], df["q1"][i] - 1.5 * iqr[i]) for i in range(df.height)],
            dtype=pl.Float64,
        )
        upper_fence: pl.Series = pl.Series(
            name="upperfence",
            values=[min(df["max"][i], df["q3"][i] + 1.5 * iqr[i]) for i in range(df.height)],
            dtype=pl.Float64,
        )
        return df.with_columns(
            lower_fence,
            upper_fence,
        )


class BoxplotChartStrategy(AbstractBoxplotChartStrategy):
    """
    Boxplot chart figure strategy.
    """

    _variables: ChartVariablesMap = {
        "category": CategoricalVariable(
            name="Category",
            description=(
                "Categorical variable(s) for the boxplot chart. "
                "Each category corresponds to a boxplot. "
                "If several variables are provided, they will be concatenated in a cartesian "
                "product with all of their values to form the final categories."
            ),
            map_to="category_variables",
            multicategorical=True,
        ),
        "value": NonDerivedNumericalVariable(
            name="Value",
            description=("A numerical variable to extract the boxplot statistics from for each category."),
            map_to="value_variable",
        ),
    }
    _options: ChartOptionsMap = {
        "horizontal": BooleanOption(
            name="Horizontal bars",
            description="When enabled, bars are rendered horizontally instead of vertically.",
            map_to="horizontal",
            default_value=False,
        ),
        "overall_category_name": StringOption(
            name="Overall category name",
            description=(
                "When set, an additional boxplot representing the overall distribution "
                "of the values (regardless of category) is displayed under the category named here."
            ),
            map_to="overall_category_name",
            default_value="",
        ),
        "display_mean": BooleanOption(
            name="Display mean",
            description="Whether to display the mean value in each boxplot.",
            map_to="display_mean",
            default_value=False,
        ),
        "display_sd": BooleanOption(
            name="Display standard deviation",
            description=("Whether to display the standard deviation in each boxplot. Only applies if mean is also displayed."),
            map_to="display_sd",
            default_value=False,
        ),
    }

    def __init__(
        self,
        fields_metadata: dict[str, FieldMetadata],
        category_variables: list[str],
        value_variable: str,
        horizontal: bool = False,
        overall_category_name: str = "",
        display_mean: bool = False,
        display_sd: bool = False,
        color_rules: list[ColorRule] | None = None,
        title: str | None = None,
        subtitle: str | None = None,
        caption: str | None = None,
        description: str | None = None,
        storytelling_context: str | None = None,
        category_label: str | None = None,
        value_label: str | None = None,
    ) -> None:
        super().__init__(
            fields_metadata=fields_metadata,
            category_variables=category_variables,
            value_variable=value_variable,
            color_rules=color_rules,
            title=title,
            subtitle=subtitle,
            caption=caption,
            description=description,
            storytelling_context=storytelling_context,
            category_label=category_label,
            value_label=value_label,
        )

        self._options["horizontal"].validate(horizontal)
        self._options["overall_category_name"].validate(overall_category_name)
        self._options["display_mean"].validate(display_mean)
        self._options["display_sd"].validate(display_sd)

        self._horizontal = horizontal
        self._overall_category_name = overall_category_name
        self._display_mean = display_mean
        self._display_sd = display_sd & self._display_mean

    @classmethod
    def name(cls) -> str:
        return "Boxplot Chart"

    @classmethod
    def description(cls) -> str:
        return (
            "A boxplot chart represents statistical distribution of samples segregated by "
            "categories. Each boxplot shows the first quartile (Q1), median (Q2), and third "
            "quartile (Q3) of the samples, as well as the computed fences (whiskers) at 1.5 times "
            "the interquartile range (IQR) from the mean."
        )

    def get_query_params(self) -> dict[str, QueryParams]:
        params = super().get_query_params()
        if self._overall_category_name:
            params["overall"] = QueryParams(
                group_fields=[],
                aggregations=[
                    TherismosAggregation(
                        id="q1",
                        function=AggregationFunction.Q1,
                        field=self._value_variable,
                    ),
                    TherismosAggregation(
                        id="median",
                        function=AggregationFunction.MEDIAN,
                        field=self._value_variable,
                    ),
                    TherismosAggregation(
                        id="q3",
                        function=AggregationFunction.Q3,
                        field=self._value_variable,
                    ),
                    TherismosAggregation(
                        id="mean",
                        function=AggregationFunction.AVERAGE,
                        field=self._value_variable,
                    ),
                    TherismosAggregation(
                        id="sd",
                        function=AggregationFunction.STDDEV,
                        field=self._value_variable,
                    ),
                    TherismosAggregation(
                        id="min",
                        function=AggregationFunction.MIN,
                        field=self._value_variable,
                    ),
                    TherismosAggregation(
                        id="max",
                        function=AggregationFunction.MAX,
                        field=self._value_variable,
                    ),
                ],
                dataframe_schema=pl.Schema(
                    {
                        "q1": pl.Float64,
                        "median": pl.Float64,
                        "q3": pl.Float64,
                        "mean": pl.Float64,
                        "sd": pl.Float64,
                        "min": pl.Float64,
                        "max": pl.Float64,
                    }
                ),
            )
        return params

    def _prepare_plot_data(
        self,
        dfs: dict[str, pl.DataFrame],
        tr: Translations,
        theme: Theme,
    ) -> dict[str, Any]:
        df: pl.DataFrame = self._compute_fences(dfs["main"])
        for category_variable in reversed(self._category_variables):
            df = self._sort_and_translate(
                df,
                category_variable,
                self._fields_metadata[category_variable],
                tr,
                sort_asc=True,
            )
        df, alias = self._concatenate_variables(df, self._category_variables)
        if self._overall_category_name:
            overall_df: pl.DataFrame = self._compute_fences(dfs["overall"])
            overall_df = overall_df.with_columns(
                pl.lit(self._overall_category_name).cast(pl.Utf8).alias(alias),
            )
            df = pl.concat(
                [
                    df,
                    overall_df.select(df.columns),
                ],
                how="vertical",
            )

        boxplot_colors: HSLColorSequence
        if self._color_rules:
            boxplot_colors = [theme.get_color("primary_color_normal")] * df.height
            for color_rule in self._color_rules:
                selector = color_rule.selector.selector(df, alias, "mean")
                assigned_colors = color_rule.assigner.assign_colors(df, alias, "mean")
                boxplot_colors = Palette.mix_color_sequences(
                    boxplot_colors,
                    assigned_colors,
                    selector,
                    indexing=Palette.MixIndexing.BY_POSITION,
                )
        else:
            boxplot_colors = ColorAssignerByCategory(
                anchor_colors=theme.get_default_color_sequence(),
            ).assign_colors(df, alias, "mean")

        return {
            "categories": df[alias].to_list(),
            "statistics": df.select(["q1", "median", "q3", "mean", "sd", "min", "max", "lowerfence", "upperfence"]).to_dicts(),
            "colors": [c.to_css_hex() for c in boxplot_colors],
            "_df": df,
            "_alias": alias,
        }

    def _render_from_payload(self, payload: dict[str, Any]) -> go.Figure:
        df = payload["_df"]
        alias = payload["_alias"]
        colors = payload["colors"]
        data = []
        for i, row in enumerate(df.iter_rows(named=True)):
            box_kwargs: dict[str, Any] = dict(
                name=row[alias],
                q1=[row["q1"]],
                median=[row["median"]],
                q3=[row["q3"]],
                lowerfence=[row["lowerfence"]],
                upperfence=[row["upperfence"]],
                boxpoints="outliers",
                marker=dict(color=colors[i]),
            )
            if self._display_mean:
                box_kwargs["mean"] = [row["mean"]]
                if self._display_sd:
                    box_kwargs["sd"] = [row["sd"]]
            if self._horizontal:
                data.append(go.Box(y=[row[alias]], orientation="h", **box_kwargs))
            else:
                data.append(go.Box(x=[row[alias]], **box_kwargs))
        fig = go.Figure(data=data)
        effective_title = payload["title"] or (f"Boxplot of {self._value_variable} by {', '.join(self._category_variables)}")
        effective_category = self._category_label or ", ".join(self._category_variables)
        effective_value = self._value_label or self._append_unit(self._value_variable, self._value_variable)
        if self._horizontal:
            fig.update_layout(xaxis_title=effective_value, yaxis_title=effective_category)
        else:
            fig.update_layout(xaxis_title=effective_category, yaxis_title=effective_value)
        self._apply_title(fig, effective_title, payload.get("subtitle"))
        return fig

    @classmethod
    def supported_color_selectors(cls) -> list[type[ColorSelectorStrategy]]:
        return [
            ColorSelectorStrategy,
            ColorSelectByCategoryName,
            ColorSelectValuesBelow,
            ColorSelectValuesAbove,
            ColorSelectMaxValue,
            ColorSelectMinValue,
            ColorSelectValuesBelowMean,
            ColorSelectValuesAboveMean,
        ]

    @classmethod
    def supported_color_assigners(cls) -> list[type[ColorAssignerStrategy]]:
        return [
            ColorAssignerByCategory,
            ColorAssignerByValue,
        ]


__all__ = [
    "AbstractBoxplotChartStrategy",
    "BoxplotChartStrategy",
]



import polars as pl
import pytest
from polychromos import HSLColor

from orama.color import (
    ColorAssignerByCategory,
    ColorAssignerByGroup,
    ColorAssignerByValue,
    ColorRule,
    ColorSelectByCategoryName,
    ColorSelectMaxValue,
    ColorSelectMinValue,
    ColorSelectorStrategy,
    ColorSelectValuesAbove,
    ColorSelectValuesAboveMean,
    ColorSelectValuesBelow,
    ColorSelectValuesBelowMean,
)

_DF = pl.DataFrame({"cat": ["A", "B", "C"], "val": [10.0, 50.0, 30.0]})
_DF_MEAN = pl.DataFrame(
    {
        "cat": ["A", "B", "C"],
        "val": [10.0, 50.0, 30.0],
        "__total_mean__": [30.0, 30.0, 30.0],
    }
)
_EMPTY_DF = pl.DataFrame({"cat": [], "val": []})


@pytest.mark.parametrize(
    "selector,expected",
    [
        (ColorSelectorStrategy(), [True, True, True]),
        (ColorSelectByCategoryName(["A", "C"]), [True, False, True]),
        (ColorSelectValuesBelow(35.0), [True, False, True]),
        (ColorSelectValuesAbove(25.0), [False, True, True]),
        (ColorSelectMaxValue(), [False, True, False]),
        (ColorSelectMinValue(), [True, False, False]),
    ],
)
def test_color_selector_returns_expected(selector: ColorSelectorStrategy, expected: list[bool]) -> None:
    result = selector.selector(_DF, "cat", "val")
    assert result == expected


def test_color_select_max_value_empty_df() -> None:
    result = ColorSelectMaxValue().selector(_EMPTY_DF, "cat", "val")
    assert result == []


def test_color_select_min_value_empty_df() -> None:
    result = ColorSelectMinValue().selector(_EMPTY_DF, "cat", "val")
    assert result == []


def test_color_select_values_below_mean() -> None:
    result = ColorSelectValuesBelowMean().selector(_DF_MEAN, "cat", "val")
    assert result == [True, False, True]


def test_color_select_values_above_mean() -> None:
    result = ColorSelectValuesAboveMean().selector(_DF_MEAN, "cat", "val")
    assert result == [False, True, True]


def test_color_assigner_strategy_empty_anchor_colors_raises() -> None:
    with pytest.raises(ValueError):
        ColorAssignerByCategory(anchor_colors=[])


def test_color_assigner_strategy_anchor_colors_property(anchor_colors: list[HSLColor]) -> None:
    assigner = ColorAssignerByCategory(anchor_colors=anchor_colors)
    assert assigner.anchor_colors == anchor_colors


def test_color_assigner_strategy_continuous_is_false(anchor_colors: list[HSLColor]) -> None:
    assigner = ColorAssignerByCategory(anchor_colors=anchor_colors)
    assert assigner.continuous is False


def test_color_assigner_by_category_assign_colors(anchor_colors: list[HSLColor]) -> None:
    assigner = ColorAssignerByCategory(anchor_colors=anchor_colors)
    colors = assigner.assign_colors(_DF, "cat", "val")
    assert len(colors) == len(_DF["cat"])


def test_color_assigner_by_category_continuous(anchor_colors: list[HSLColor]) -> None:
    assigner = ColorAssignerByCategory(anchor_colors=anchor_colors)
    assert assigner.continuous is False


def test_color_assigner_by_group_no_group_var_raises(anchor_colors: list[HSLColor]) -> None:
    assigner = ColorAssignerByGroup(anchor_colors=anchor_colors)
    with pytest.raises(ValueError):
        assigner.assign_colors(_DF, "cat", "val", group_var=None)


def test_color_assigner_by_group_assign_colors(anchor_colors: list[HSLColor]) -> None:
    df = pl.DataFrame({"cat": ["A", "A", "B"], "val": [1.0, 2.0, 3.0], "group": ["G1", "G1", "G2"]})
    assigner = ColorAssignerByGroup(anchor_colors=anchor_colors)
    colors = assigner.assign_colors(df, "cat", "val", group_var="group")
    assert len(colors) == len(df["group"].unique())


def test_color_assigner_by_value_continuous(anchor_colors: list[HSLColor]) -> None:
    assigner = ColorAssignerByValue(anchor_colors=anchor_colors)
    assert assigner.continuous is True


def test_color_assigner_by_value_properties(anchor_colors: list[HSLColor]) -> None:
    assigner = ColorAssignerByValue(
        anchor_colors=anchor_colors,
        range_min=0.0,
        range_max=100.0,
        range_center_at=50.0,
    )
    assert assigner.range_min == 0.0
    assert assigner.range_max == 100.0
    assert assigner.range_center_at == 50.0


def test_color_assigner_by_value_properties_none_defaults(anchor_colors: list[HSLColor]) -> None:
    assigner = ColorAssignerByValue(anchor_colors=anchor_colors)
    assert assigner.range_min is None
    assert assigner.range_max is None
    assert assigner.range_center_at is None


def test_color_assigner_by_value_assign_colors_length(anchor_colors: list[HSLColor]) -> None:
    assigner = ColorAssignerByValue(anchor_colors=anchor_colors)
    colors = assigner.assign_colors(_DF, "cat", "val")
    assert len(colors) == _DF.height


def test_color_rule_default_selector(anchor_colors: list[HSLColor]) -> None:
    assigner = ColorAssignerByCategory(anchor_colors=anchor_colors)
    rule = ColorRule(assigner=assigner)
    assert isinstance(rule.selector, ColorSelectorStrategy)


def test_color_rule_properties(anchor_colors: list[HSLColor]) -> None:
    assigner = ColorAssignerByCategory(anchor_colors=anchor_colors)
    selector = ColorSelectByCategoryName(["A"])
    rule = ColorRule(assigner=assigner, selector=selector)
    assert rule.selector is selector
    assert rule.assigner is assigner


@pytest.mark.parametrize(
    "selector,expected",
    [
        (ColorSelectorStrategy(), "selecting all entries"),
        (ColorSelectByCategoryName(["A", "B"]), "selecting entries in categories ['A', 'B']"),
        (ColorSelectValuesBelow(10.0), "selecting entries with a value below 10.0"),
        (ColorSelectValuesAbove(5.0), "selecting entries with a value above 5.0"),
        (ColorSelectMaxValue(), "selecting the entry with the highest value"),
        (ColorSelectMinValue(), "selecting the entry with the lowest value"),
        (ColorSelectValuesBelowMean(), "selecting entries with a value below the mean"),
        (ColorSelectValuesAboveMean(), "selecting entries with a value above the mean"),
    ],
)
def test_color_selector_explain(selector: ColorSelectorStrategy, expected: str) -> None:
    assert selector.explain() == expected


def test_color_assigner_by_category_explain(anchor_colors: list[HSLColor]) -> None:
    assigner = ColorAssignerByCategory(anchor_colors=anchor_colors)
    result = assigner.explain()
    assert result.startswith("colorize them each in a distinct color from")
    assert anchor_colors[0].to_css_hex() in result


def test_color_assigner_by_group_explain(anchor_colors: list[HSLColor]) -> None:
    assigner = ColorAssignerByGroup(anchor_colors=anchor_colors)
    result = assigner.explain()
    assert result.startswith("colorize them each in a distinct color by group from")
    assert anchor_colors[0].to_css_hex() in result


def test_color_assigner_by_value_explain_no_range(anchor_colors: list[HSLColor]) -> None:
    assigner = ColorAssignerByValue(anchor_colors=anchor_colors)
    result = assigner.explain()
    assert result.startswith("colorize them by their value in colors from")
    assert anchor_colors[0].to_css_hex() in result
    assert "restricted" not in result
    assert "centered" not in result


def test_color_assigner_by_value_explain_with_range(anchor_colors: list[HSLColor]) -> None:
    assigner = ColorAssignerByValue(anchor_colors=anchor_colors, range_min=0.0, range_max=100.0)
    result = assigner.explain()
    assert "restricted to values 0.0\u2013100.0" in result


def test_color_assigner_by_value_explain_with_center(anchor_colors: list[HSLColor]) -> None:
    assigner = ColorAssignerByValue(anchor_colors=anchor_colors, range_center_at=50.0)
    result = assigner.explain()
    assert "centered at 50.0" in result


def test_color_rule_explain(anchor_colors: list[HSLColor]) -> None:
    selector = ColorSelectValuesAbove(12.0)
    assigner = ColorAssignerByCategory(anchor_colors=anchor_colors)
    rule = ColorRule(assigner=assigner, selector=selector)
    result = rule.explain()
    assert result.startswith("selecting entries with a value above 12.0")
    assert "colorize them" in result


def test_color_rule_explain_default_selector(anchor_colors: list[HSLColor]) -> None:
    assigner = ColorAssignerByCategory(anchor_colors=anchor_colors)
    rule = ColorRule(assigner=assigner)
    result = rule.explain()
    assert result.startswith("selecting all entries")
    assert "colorize them" in result



"""
Figure strategy implementations for orama.
"""

from orama.strategies.bar import AbstractBarChartStrategy, BarChartStrategy
from orama.strategies.base import FigureStrategy, FigureStrategyDescription, FigureView
from orama.strategies.boxplot import AbstractBoxplotChartStrategy, BoxplotChartStrategy
from orama.strategies.categorical_line import CategoricalLineChartStrategy
from orama.strategies.choropleth_map import ChoroplethMapChartStrategy
from orama.strategies.grouped_bar import GroupedBarChartStrategy
from orama.strategies.grouped_boxplot import GroupedBoxplotChartStrategy
from orama.strategies.heatmap import HeatMapChartStrategy
from orama.strategies.line import LineChartStrategy
from orama.strategies.pie import PieChartStrategy
from orama.strategies.sunburst import SunburstChartStrategy

__all__ = [
    "FigureStrategy",
    "FigureStrategyDescription",
    "FigureView",
    "AbstractBarChartStrategy",
    "BarChartStrategy",
    "GroupedBarChartStrategy",
    "LineChartStrategy",
    "CategoricalLineChartStrategy",
    "PieChartStrategy",
    "SunburstChartStrategy",
    "HeatMapChartStrategy",
    "AbstractBoxplotChartStrategy",
    "BoxplotChartStrategy",
    "GroupedBoxplotChartStrategy",
    "ChoroplethMapChartStrategy",
]



"""
Unit tests for BoxplotChartStrategy._prepare_plot_data().
"""

from typing import Any
from unittest.mock import MagicMock

import polars as pl
import pytest
from fields_metadata import FieldMetadata
from polychromos import HSLColor

from orama.color import (
    ColorAssignerByCategory,
    ColorRule,
    ColorSelectorStrategy,
)
from orama.strategies.boxplot import BoxplotChartStrategy


def _cat_meta() -> MagicMock:
    m = MagicMock(spec=FieldMetadata)
    m.categorical = True
    m.numeric = False
    m.derived = False
    m.effective_type = str
    m.extra = {}
    return m


def _num_meta() -> MagicMock:
    m = MagicMock(spec=FieldMetadata)
    m.categorical = False
    m.numeric = True
    m.derived = False
    m.effective_type = float
    m.extra = {}
    return m


def _make_boxplot(**kwargs: Any) -> BoxplotChartStrategy:
    return BoxplotChartStrategy(
        fields_metadata={"cat": _cat_meta(), "val": _num_meta()},
        category_variables=["cat"],
        value_variable="val",
        **kwargs,
    )


def _stats_df(categories: list[str]) -> pl.DataFrame:
    """Build a minimal statistics DataFrame as would come from a query."""
    n = len(categories)
    return pl.DataFrame(
        {
            "cat": categories,
            "q1": [1.0] * n,
            "median": [2.0] * n,
            "q3": [3.0] * n,
            "mean": [2.0] * n,
            "sd": [0.5] * n,
            "min": [0.0] * n,
            "max": [4.0] * n,
        }
    )


@pytest.fixture
def stats_df() -> pl.DataFrame:
    return _stats_df(["A", "B"])


def test_prepare_plot_data_returns_categories_statistics_colors(
    stats_df: pl.DataFrame,
    tr: Any,
    theme: Any,
) -> None:
    boxplot = _make_boxplot()
    result = boxplot._prepare_plot_data({"main": stats_df}, tr, theme)
    assert result["categories"] == ["A", "B"]
    assert len(result["statistics"]) == 2
    assert len(result["colors"]) == 2
    assert all(c.startswith("#") for c in result["colors"])
    # _compute_fences must have added fences to each statistics dict
    for stat in result["statistics"]:
        assert "lowerfence" in stat
        assert "upperfence" in stat


def test_prepare_plot_data_overall_category_appended(
    stats_df: pl.DataFrame,
    tr: Any,
    theme: Any,
) -> None:
    overall_df = _stats_df([""])
    # overall query has no category columns — just the stats
    overall_df = overall_df.drop("cat")
    boxplot = _make_boxplot(overall_category_name="Total")
    result = boxplot._prepare_plot_data({"main": stats_df, "overall": overall_df}, tr, theme)
    assert result["categories"][-1] == "Total"
    assert len(result["categories"]) == 3


def test_prepare_plot_data_color_rules_applied(
    stats_df: pl.DataFrame,
    anchor_colors: list[HSLColor],
    tr: Any,
    theme: Any,
) -> None:
    rule = ColorRule(
        assigner=ColorAssignerByCategory(anchor_colors=anchor_colors),
        selector=ColorSelectorStrategy(),
    )
    boxplot = _make_boxplot(color_rules=[rule])
    result = boxplot._prepare_plot_data({"main": stats_df}, tr, theme)
    assert len(result["colors"]) == 2
    assert all(c.startswith("#") for c in result["colors"])



"""
Unit tests for GroupedBoxplotChartStrategy._prepare_plot_data().
"""

from typing import Any
from unittest.mock import MagicMock

import polars as pl
import pytest
from fields_metadata import FieldMetadata

from orama.strategies.grouped_boxplot import GroupedBoxplotChartStrategy


def _cat_meta() -> MagicMock:
    m = MagicMock(spec=FieldMetadata)
    m.categorical = True
    m.numeric = False
    m.derived = False
    m.effective_type = str
    m.extra = {}
    return m


def _num_meta() -> MagicMock:
    m = MagicMock(spec=FieldMetadata)
    m.categorical = False
    m.numeric = True
    m.derived = False
    m.effective_type = float
    m.extra = {}
    return m


def _make_grouped_boxplot(**kwargs: Any) -> GroupedBoxplotChartStrategy:
    return GroupedBoxplotChartStrategy(
        fields_metadata={"cat": _cat_meta(), "val": _num_meta(), "grp": _cat_meta()},
        category_variables=["cat"],
        value_variable="val",
        group_variable="grp",
        **kwargs,
    )


@pytest.fixture
def grouped_stats_df() -> pl.DataFrame:
    return pl.DataFrame(
        {
            "cat": ["A", "A", "B", "B"],
            "grp": ["X", "Y", "X", "Y"],
            "q1": [1.0, 1.5, 2.0, 2.5],
            "median": [2.0, 2.5, 3.0, 3.5],
            "q3": [3.0, 3.5, 4.0, 4.5],
            "mean": [2.0, 2.5, 3.0, 3.5],
            "sd": [0.5, 0.5, 0.5, 0.5],
            "min": [0.0, 0.5, 1.0, 1.5],
            "max": [4.0, 4.5, 5.0, 5.5],
        }
    )


def test_prepare_plot_data_returns_categories_groups_color_map(
    grouped_stats_df: pl.DataFrame,
    tr: Any,
    theme: Any,
) -> None:
    instance = _make_grouped_boxplot()
    result = instance._prepare_plot_data({"main": grouped_stats_df}, tr, theme)
    assert isinstance(result["categories"], list)
    assert isinstance(result["groups"], list)
    assert isinstance(result["color_map"], dict)
    assert "_group_colors" in result


def test_prepare_plot_data_color_map_keyed_by_group_values(
    grouped_stats_df: pl.DataFrame,
    tr: Any,
    theme: Any,
) -> None:
    instance = _make_grouped_boxplot()
    result = instance._prepare_plot_data({"main": grouped_stats_df}, tr, theme)
    assert set(result["color_map"].keys()) == {"X", "Y"}
    for v in result["color_map"].values():
        assert v.startswith("#")



"""
Defines classes and data structures for figure variables.
"""

import datetime
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any

from fields_metadata import FieldMetadata


@dataclass(frozen=True)
class Aggregation:
    """
    Encapsulates the parameters of an aggregation variable.

    :param name: Output alias for the aggregated value.
    :type name: str
    :param function: Aggregation function identifier string.
    :type function: str
    :param field: Source field to aggregate. ``None`` for count-only aggregations.
    :type field: str | None
    """

    name: str
    function: str
    field: str | None = None

    @property
    def default_label(self) -> str:
        """
        Returns a human-readable label derived from the aggregation function and field.

        :return: Descriptive label such as ``"Average of revenue"`` or ``"Count"``.
        :rtype: str
        """
        fn_label = self.function.replace("_", " ").title()
        if self.field is not None:
            return f"{fn_label} of {self.field}"
        return fn_label


class ChartVariable(ABC):
    """
    Abstract base class for chart variables.
    """

    def __init__(
        self,
        name: str,
        description: str,
        map_to: str,
        icon: str | None = None,
    ) -> None:
        self._name = name
        self._description = description
        self._map_to = map_to
        self._icon = icon

    @abstractmethod
    def candidate_fields(
        self,
        fields_metadata: dict[str, FieldMetadata],
    ) -> list[str]:
        """
        Returns a list of candidate fields for this variable.

        :param fields_metadata: Mapping of field paths to their metadata.
        :type fields_metadata: dict[str, FieldMetadata]
        :return: A list of candidate field names.
        :rtype: list[str]
        """

    def validate_fields(
        self,
        fields_metadata: dict[str, FieldMetadata],
        *fields: str,
    ) -> None:
        """
        Validates a number of fields according to the variable's parameters.

        :param fields_metadata: Mapping of field paths to their metadata.
        :type fields_metadata: dict[str, FieldMetadata]
        :raises ValueError: When any of the fields are not valid candidates.
        """
        candidates = self.candidate_fields(fields_metadata)
        for field in fields:
            if field not in candidates:
                raise ValueError(f"Field {field} is not a valid candidate field for variable {self._name}")

    def describe(
        self,
        fields_metadata: dict[str, FieldMetadata],
    ) -> dict[str, Any]:
        """
        Provides a descriptive structure of the variable.

        :param fields_metadata: Mapping of field paths to their metadata.
        :type fields_metadata: dict[str, FieldMetadata]
        :return: A dictionary describing the variable.
        :rtype: dict[str, Any]
        """
        return {
            "type": self.__class__,
            "name": self._name,
            "description": self._description,
            "map_to": self._map_to,
            "icon": self._icon,
            "candidate_fields": self.candidate_fields(fields_metadata),
        }


class CategoricalVariable(ChartVariable):
    """
    Represents a (multi)categorical variable for charts.
    """

    def __init__(
        self,
        name: str,
        description: str,
        map_to: str,
        multicategorical: bool = False,
        ordered: bool = False,
        optional: bool = False,
        icon: str | None = None,
    ) -> None:
        """
        Constructor method.

        :param name: The name of the variable.
        :type name: str
        :param description: The description of the variable.
        :type description: str
        :param map_to: The internal parameter to map the variable to.
        :type map_to: str
        :param multicategorical: When ``True``, treats the variable as multicategorical.
        :type multicategorical: bool, optional
        :param ordered: When ``True``, renders as an ordered FieldList of single-select entries.
        :type ordered: bool, optional
        :param optional: When ``True``, prepends a "— none —" empty choice to the select field.
        :type optional: bool, optional
        :param icon: An optional SVG icon string. Defaults to ``None``.
        :type icon: str | None, optional
        """
        super().__init__(
            name=name,
            description=description,
            map_to=map_to,
            icon=icon,
        )
        self._multicategorical: bool = multicategorical
        self._ordered: bool = ordered
        self._optional: bool = optional

    def candidate_fields(
        self,
        fields_metadata: dict[str, FieldMetadata],
    ) -> list[str]:
        return [field for field, metadata in fields_metadata.items() if metadata.categorical]

    def validate_fields(
        self,
        fields_metadata: dict[str, FieldMetadata],
        *fields: str,
    ) -> None:
        """
        Validates a number of fields according to the variable's parameters.

        :param fields_metadata: Mapping of field paths to their metadata.
        :type fields_metadata: dict[str, FieldMetadata]
        :raises ValueError: When more than one field is provided and not multicategorical,
            or when a field is not a valid candidate.
        """
        if len(fields) != 1 and not self._multicategorical:
            raise ValueError("This variable only accepts a single field")
        super().validate_fields(fields_metadata, *fields)

    def describe(
        self,
        fields_metadata: dict[str, FieldMetadata],
    ) -> dict[str, Any]:
        base_description = super().describe(fields_metadata)
        base_description["multicategorical"] = self._multicategorical
        base_description["ordered"] = self._ordered
        base_description["optional"] = self._optional
        return base_description


class NumericalVariable(ChartVariable):
    """
    Represents a numerical variable for charts.
    """

    def __init__(
        self,
        name: str,
        description: str,
        map_to: str,
        icon: str | None = None,
    ) -> None:
        """
        Constructor method.

        :param name: The name of the variable.
        :type name: str
        :param description: The description of the variable.
        :type description: str
        :param map_to: The internal parameter to map the variable to.
        :type map_to: str
        :param icon: An optional SVG icon string. Defaults to ``None``.
        :type icon: str | None, optional
        """
        super().__init__(
            name=name,
            description=description,
            map_to=map_to,
            icon=icon,
        )

    def candidate_fields(
        self,
        fields_metadata: dict[str, FieldMetadata],
    ) -> list[str]:
        return [field for field, metadata in fields_metadata.items() if metadata.numeric]

    def validate_fields(
        self,
        fields_metadata: dict[str, FieldMetadata],
        *fields: str,
    ) -> None:
        """
        Validates a number of fields according to the variable's parameters.

        :param fields_metadata: Mapping of field paths to their metadata.
        :type fields_metadata: dict[str, FieldMetadata]
        :raises ValueError: When more than one field is provided.
        """
        if len(fields) != 1:
            raise ValueError("This variable only accepts a single field")
        super().validate_fields(fields_metadata, *fields)


class NonDerivedNumericalVariable(NumericalVariable):
    """
    Represents a non-derived numerical variable for charts.
    """

    def __init__(
        self,
        name: str,
        description: str,
        map_to: str,
        icon: str | None = None,
    ) -> None:
        """
        Constructor method.

        :param name: The name of the variable.
        :type name: str
        :param description: The description of the variable.
        :type description: str
        :param map_to: The internal parameter to map the variable to.
        :type map_to: str
        :param icon: An optional SVG icon string. Defaults to ``None``.
        :type icon: str | None, optional
        """
        super().__init__(
            name=name,
            description=description,
            map_to=map_to,
            icon=icon,
        )

    def candidate_fields(
        self,
        fields_metadata: dict[str, FieldMetadata],
    ) -> list[str]:
        return [field for field, metadata in fields_metadata.items() if metadata.numeric and not metadata.derived]


class NumericalAggregationVariable(NonDerivedNumericalVariable):
    """
    Represents a numerical aggregation variable for charts.
    """

    def __init__(
        self,
        name: str,
        description: str,
        map_to: str,
        icon: str | None = None,
    ) -> None:
        """
        Constructor method.

        :param name: The name of the variable.
        :type name: str
        :param description: The description of the variable.
        :type description: str
        :param map_to: The internal parameter to map the variable to.
        :type map_to: str
        :param icon: An optional SVG icon string. Defaults to ``None``.
        :type icon: str | None, optional
        """
        super().__init__(
            name=name,
            description=description,
            map_to=map_to,
            icon=icon,
        )


class DateTimeVariable(ChartVariable):
    """
    Represents a datetime variable for charts.
    """

    def __init__(
        self,
        name: str,
        description: str,
        map_to: str,
        icon: str | None = None,
    ) -> None:
        """
        Constructor method.

        :param name: The name of the variable.
        :type name: str
        :param description: The description of the variable.
        :type description: str
        :param map_to: The internal parameter to map the variable to.
        :type map_to: str
        :param icon: An optional SVG icon string. Defaults to ``None``.
        :type icon: str | None, optional
        """
        super().__init__(
            name=name,
            description=description,
            map_to=map_to,
            icon=icon,
        )

    def candidate_fields(
        self,
        fields_metadata: dict[str, FieldMetadata],
    ) -> list[str]:
        return [
            field
            for field, metadata in fields_metadata.items()
            if metadata.effective_type is not None and issubclass(metadata.effective_type, (datetime.date, datetime.datetime))
        ]

    def validate_fields(
        self,
        fields_metadata: dict[str, FieldMetadata],
        *fields: str,
    ) -> None:
        """
        Validates a number of fields according to the variable's parameters.

        :param fields_metadata: Mapping of field paths to their metadata.
        :type fields_metadata: dict[str, FieldMetadata]
        :raises ValueError: When more than one field is provided.
        """
        if len(fields) != 1:
            raise ValueError("This variable only accepts a single field")
        super().validate_fields(fields_metadata, *fields)


class CountryCodeVariable(ChartVariable):
    """
    Represents a variable accepting only fields tagged as ISO 3166-1 alpha-2 country codes.

    Fields qualify when their metadata carries
    ``extra["suggested_validation"] == "iso_a2"``.
    """

    def candidate_fields(self, fields_metadata: dict[str, FieldMetadata]) -> list[str]:
        """
        Returns fields whose metadata carries ``suggested_validation == "iso_a2"``.

        :param fields_metadata: Mapping of field paths to their metadata.
        :type fields_metadata: dict[str, FieldMetadata]
        :return: A list of candidate field names.
        :rtype: list[str]
        """
        return [field for field, metadata in fields_metadata.items() if metadata.extra.get("suggested_validation") == "iso_a2"]

    def validate_fields(self, fields_metadata: dict[str, FieldMetadata], *fields: str) -> None:
        """
        Validates that exactly one field is provided and that it is a valid candidate.

        :param fields_metadata: Mapping of field paths to their metadata.
        :type fields_metadata: dict[str, FieldMetadata]
        :raises ValueError: When the number of fields is not exactly one,
            or the field is not a valid candidate.
        """
        if len(fields) != 1:
            raise ValueError("This variable only accepts a single field")
        super().validate_fields(fields_metadata, *fields)


ChartVariablesMap = dict[str, ChartVariable]


__all__ = [
    "Aggregation",
    "ChartVariable",
    "CategoricalVariable",
    "CountryCodeVariable",
    "NumericalVariable",
    "NonDerivedNumericalVariable",
    "NumericalAggregationVariable",
    "DateTimeVariable",
    "ChartVariablesMap",
]



"""
Line chart figure strategy implementation.
"""

from itertools import groupby
from typing import Any

import plotly.graph_objects as go
import polars as pl
from babel.support import Translations
from fields_metadata import FieldMetadata
from plotly.subplots import make_subplots
from polychromos.palette import HSLColorSequence
from therismos.grouping import Aggregation as TherismosAggregation
from therismos.grouping import AggregationFunction

from orama.color import (
    ColorAssignerByCategory,
    ColorAssignerStrategy,
    ColorRule,
    ColorSelectByCategoryName,
    ColorSelectorStrategy,
)
from orama.enums import LineInterpolation, LinePlacement, LineSortOrder
from orama.options import (
    ChartOptionsMap,
    FloatOption,
    IntOption,
    LineInterpolationOption,
    LinePlacementOption,
    LineSortOrderOption,
)
from orama.query import QueryParams
from orama.strategies.base import FigureStrategy, _hex_to_rgb
from orama.theme import Theme
from orama.variables import (
    Aggregation,
    CategoricalVariable,
    ChartVariablesMap,
    DateTimeVariable,
    NumericalAggregationVariable,
)


class LineChartStrategy(FigureStrategy):
    """
    Line chart figure strategy.
    """

    _variables: ChartVariablesMap = {
        "date": DateTimeVariable(
            name="Date",
            description="Calendar variable. Documents will be aggregated by year and month.",
            map_to="date_variable",
        ),
        "category": CategoricalVariable(
            name="Category",
            description=("Categorical variable for the line chart. Each category corresponds to a line."),
            map_to="category_variable",
            multicategorical=False,
        ),
        "value": NumericalAggregationVariable(
            name="Value",
            description=("An aggregation of a numerical variable to produce the final value at each point in time."),
            map_to="aggregation",
        ),
    }
    _options: ChartOptionsMap = {
        "min_value": FloatOption(
            name="Minimum value",
            description="Minimum value for the line chart value axis.",
            map_to="min_value",
            default_value=None,
        ),
        "max_value": FloatOption(
            name="Maximum value",
            description="Maximum value for the line chart value axis.",
            map_to="max_value",
            default_value=None,
        ),
        "fill_opacity": FloatOption(
            name="Fill opacity",
            description="Opacity of the filled area under each line. 0 disables the fill.",
            map_to="fill_opacity",
            default_value=0.0,
            min_value=0.0,
            max_value=1.0,
        ),
        "line_width": IntOption(
            name="Line width",
            description="Width of each line in pixels.",
            map_to="line_width",
            default_value=3,
            min_value=1,
        ),
        "placement": LinePlacementOption(
            name="Placement",
            description="How the series lines/areas are positioned relative to each other.",
            map_to="placement",
            default_value=LinePlacement.OVERLAPPING,
        ),
        "interpolation": LineInterpolationOption(
            name="Interpolation",
            description="Line interpolation mode between data points.",
            map_to="interpolation",
            default_value=LineInterpolation.LINEAR,
        ),
        "sort_order": LineSortOrderOption(
            name="Sort series",
            description="Order in which the category lines are drawn.",
            map_to="sort_order",
            default_value=LineSortOrder.NONE,
        ),
    }

    def __init__(
        self,
        fields_metadata: dict[str, FieldMetadata],
        date_variable: str,
        aggregation: Aggregation,
        category_variable: str | None = None,
        min_value: float | None = None,
        max_value: float | None = None,
        fill_opacity: float = 0.0,
        line_width: int = 3,
        placement: LinePlacement = LinePlacement.OVERLAPPING,
        interpolation: LineInterpolation = LineInterpolation.LINEAR,
        sort_order: LineSortOrder = LineSortOrder.NONE,
        color_rules: list[ColorRule] | None = None,
        title: str | None = None,
        subtitle: str | None = None,
        caption: str | None = None,
        description: str | None = None,
        storytelling_context: str | None = None,
        date_label: str | None = None,
        value_label: str | None = None,
        category_label: str | None = None,
    ) -> None:
        super().__init__(
            fields_metadata=fields_metadata,
            color_rules=color_rules,
            title=title,
            subtitle=subtitle,
            caption=caption,
            description=description,
            storytelling_context=storytelling_context,
        )

        self._variables["date"].validate_fields(fields_metadata, date_variable)
        if aggregation.field:
            self._variables["value"].validate_fields(fields_metadata, aggregation.field)
        if category_variable:
            self._variables["category"].validate_fields(fields_metadata, category_variable)

        self._date_variable = date_variable
        self._aggregation = aggregation
        self._explicit_category_variable = category_variable
        self._category_variable = category_variable or "type"

        if min_value is not None:
            self._options["min_value"].validate(min_value)
        if max_value is not None:
            self._options["max_value"].validate(max_value)
        self._options["fill_opacity"].validate(fill_opacity)
        self._options["line_width"].validate(line_width)
        self._options["placement"].validate(placement)
        self._options["interpolation"].validate(interpolation)
        self._options["sort_order"].validate(sort_order)

        self._min_value = min_value
        self._max_value = max_value
        self._fill_opacity = fill_opacity
        self._line_width = line_width
        self._placement = placement
        self._interpolation = interpolation
        self._sort_order = sort_order
        self._date_label = date_label
        self._value_label = value_label
        self._category_label = category_label

    @classmethod
    def name(cls) -> str:
        return "Line Chart"

    @classmethod
    def description(cls) -> str:
        return (
            "A line chart represents information as a series of data points connected by straight "
            "line segments, used to visualize a trend in data over intervals of time."
        )

    def get_query_params(self) -> dict[str, QueryParams]:
        return {
            "main": QueryParams(
                group_fields=[
                    f"{self._date_variable}__year",
                    f"{self._date_variable}__month",
                    self._category_variable,
                ],
                aggregations=[
                    TherismosAggregation(
                        id=self._aggregation.name,
                        function=AggregationFunction(self._aggregation.function),
                        field=self._aggregation.field,
                    ),
                ],
                dataframe_schema=pl.Schema(
                    self._get_schema_fields(),
                ),
            )
        }

    def _get_schema_fields(
        self,
        datetime_vars: list[str] | None = None,
        value_var: str | None = None,
    ) -> dict[str, type[pl.DataType]]:
        if not datetime_vars:
            datetime_vars = [
                f"{self._date_variable}__year",
                f"{self._date_variable}__month",
                self._category_variable,
            ]
        if not value_var:
            value_var = self._aggregation.name
        return {
            **{var: self._infer_column_type(self._fields_metadata, var) for var in datetime_vars},
            value_var: (pl.Int64 if self._aggregation.function == AggregationFunction.COUNT else pl.Float64),
        }

    def _build_line_dataframes(
        self,
        df: pl.DataFrame,
        tr: Translations,
    ) -> tuple[dict[str, pl.DataFrame], list[str], str]:
        """
        Build per-category DataFrames with date-range filling, null trimming, and date translation.

        :param df: Raw DataFrame from the query result.
        :type df: pl.DataFrame
        :param tr: Translations for localizing date labels.
        :type tr: Translations
        :return: Mapping of category to its processed DataFrame, ordered category list, and date alias.
        :rtype: tuple[dict[str, pl.DataFrame], list[str], str]
        """
        df = self._sort_and_translate(
            df,
            self._category_variable,
            self._fields_metadata[self._category_variable],
            tr,
            sort_asc=True,
        )
        line_dfs: dict[str, pl.DataFrame] = {}
        date_variables = (f"{self._date_variable}__year", f"{self._date_variable}__month")
        categories = [key for key, _ in groupby(df[self._category_variable].to_list())]
        alias = ""
        for category in categories:
            _df = df.filter(pl.col(self._category_variable) == category).select([*date_variables, self._aggregation.name])
            if not _df.height:
                continue
            for year in range(_df[date_variables[0]].min(), _df[date_variables[0]].max() + 1):  # type: ignore[arg-type, operator]
                for month in range(1, 13):
                    if not ((_df[date_variables[0]] == year) & (_df[date_variables[1]] == month)).any():
                        new_row = pl.DataFrame(
                            {
                                date_variables[0]: [year],
                                date_variables[1]: [month],
                                self._aggregation.name: [None],
                            }
                        )
                        _df = pl.concat([_df, new_row], how="vertical")
            _df = _df.sort(by=list(date_variables))
            _df = _df.with_row_index("__row_nr")
            agg_col = self._aggregation.name
            non_null = _df.filter(pl.col(agg_col).is_not_null())
            if non_null.height == 0:
                _df = _df.head(0)
            else:
                start = int(non_null["__row_nr"].min())  # type: ignore[arg-type]
                end = int(non_null["__row_nr"].max())  # type: ignore[arg-type]
                _df = _df.filter((pl.col("__row_nr") >= start) & (pl.col("__row_nr") <= end))
            if "__row_nr" in _df.columns:
                _df = _df.drop("__row_nr")
            for category_variable in reversed(date_variables):
                _df = self._sort_and_translate(
                    _df,
                    category_variable,
                    self._fields_metadata[category_variable],
                    tr,
                    sort_asc=True,
                )
            _df, alias = self._concatenate_variables(_df, list(date_variables))
            line_dfs[category] = _df
        return line_dfs, categories, alias

    def _compute_line_colors(
        self,
        categories_df: pl.DataFrame,
        theme: Theme,
    ) -> HSLColorSequence:
        """
        Produce a per-category color sequence via color rules or the default category assigner.

        :param categories_df: Single-column DataFrame of category values.
        :type categories_df: pl.DataFrame
        :param theme: Theme providing color sequences.
        :type theme: Theme
        :return: A color sequence with one entry per category.
        :rtype: HSLColorSequence
        """
        if self._color_rules:
            return self._apply_color_rules(categories_df, self._category_variable, self._aggregation.name, theme)
        return ColorAssignerByCategory(
            anchor_colors=theme.get_default_color_sequence(),
        ).assign_colors(categories_df, self._category_variable, self._aggregation.name)

    def _compute_value_range(
        self,
        df: pl.DataFrame,
    ) -> tuple[float, float]:
        """
        Derive the y-axis range from explicit overrides or the data extremes with a 5 % margin.

        :param df: The full (unfiltered) DataFrame.
        :type df: pl.DataFrame
        :return: ``(min_value, max_value)`` tuple for the y-axis range.
        :rtype: tuple[float, float]
        """
        min_val = (
            self._min_value
            if self._min_value is not None
            else (
                float(df[self._aggregation.name].min())  # type: ignore[arg-type]
                - 0.05 * abs(float(df[self._aggregation.name].cast(pl.Float64).min()))  # type: ignore[arg-type]
            )
        )
        max_val = (
            self._max_value
            if self._max_value is not None
            else (
                float(df[self._aggregation.name].max())  # type: ignore[arg-type]
                + 0.05 * abs(float(df[self._aggregation.name].cast(pl.Float64).max()))  # type: ignore[arg-type]
            )
        )
        return min_val, max_val

    def _sort_active_categories(
        self,
        active_categories: list[str],
        line_dfs: dict[str, pl.DataFrame],
    ) -> list[str]:
        """
        Re-order active categories according to :attr:`_sort_order`.

        :param active_categories: Categories present in ``line_dfs``.
        :type active_categories: list[str]
        :param line_dfs: Per-category DataFrames used to read first/last values.
        :type line_dfs: dict[str, pl.DataFrame]
        :return: Sorted copy of ``active_categories``.
        :rtype: list[str]
        """
        if self._sort_order == LineSortOrder.NONE:
            return active_categories
        if self._sort_order == LineSortOrder.ALPHABETICAL_ASC:
            return sorted(active_categories, key=str)
        if self._sort_order == LineSortOrder.ALPHABETICAL_DESC:
            return sorted(active_categories, key=str, reverse=True)
        if self._sort_order in (LineSortOrder.FIRST_VALUE_ASC, LineSortOrder.FIRST_VALUE_DESC):

            def _first(c: str) -> float:
                vals = line_dfs[c][self._aggregation.name].drop_nulls()
                return float(vals[0]) if vals.len() > 0 else 0.0

            return sorted(
                active_categories,
                key=_first,
                reverse=(self._sort_order == LineSortOrder.FIRST_VALUE_DESC),
            )
        if self._sort_order in (LineSortOrder.LAST_VALUE_ASC, LineSortOrder.LAST_VALUE_DESC):

            def _last(c: str) -> float:
                vals = line_dfs[c][self._aggregation.name].drop_nulls()
                return float(vals[-1]) if vals.len() > 0 else 0.0

            return sorted(
                active_categories,
                key=_last,
                reverse=(self._sort_order == LineSortOrder.LAST_VALUE_DESC),
            )
        return active_categories

    def _build_ridgeline_figure(
        self,
        active_categories: list[str],
        categories: list[str],
        line_dfs: dict[str, pl.DataFrame],
        line_colors: HSLColorSequence,
        alias: str,
        value_range: tuple[float, float],
    ) -> go.Figure:
        """
        Build a ridgeline (vertically stacked, shared x-axis) figure.

        :param active_categories: Ordered subset of categories to render.
        :type active_categories: list[str]
        :param categories: Full category list (used for color index lookup).
        :type categories: list[str]
        :param line_dfs: Per-category DataFrames.
        :type line_dfs: dict[str, pl.DataFrame]
        :param line_colors: Per-category color sequence.
        :type line_colors: HSLColorSequence
        :param alias: Date axis column name.
        :type alias: str
        :param value_range: Y-axis range applied when explicit bounds are set.
        :type value_range: tuple[float, float]
        :return: Ridgeline figure.
        :rtype: go.Figure
        """
        fig = make_subplots(
            rows=len(active_categories),
            cols=1,
            shared_xaxes=True,
            vertical_spacing=0.02,
        )
        for idx, category in enumerate(active_categories):
            i = categories.index(category)
            r, g, b = _hex_to_rgb(line_colors[i].to_css_hex())
            fill_kwargs: dict[str, Any] = {}
            if self._fill_opacity > 0:
                fill_kwargs = dict(
                    fill="tozeroy",
                    fillcolor=f"rgba({r},{g},{b},{self._fill_opacity})",
                )
            fig.add_trace(
                go.Scatter(
                    x=line_dfs[category][alias],
                    y=line_dfs[category][self._aggregation.name],
                    mode="lines",
                    line=dict(color=line_colors[i].to_css_hex(), width=self._line_width, shape=self._interpolation.value),
                    connectgaps=True,
                    name=str(category),
                    **fill_kwargs,
                ),
                row=idx + 1,
                col=1,
            )
            if self._min_value is not None or self._max_value is not None:
                fig.update_yaxes(range=value_range, row=idx + 1, col=1)
        return fig

    def _build_standard_figure(
        self,
        active_categories: list[str],
        categories: list[str],
        line_dfs: dict[str, pl.DataFrame],
        line_colors: HSLColorSequence,
        alias: str,
        value_range: tuple[float, float],
    ) -> go.Figure:
        """
        Build an overlapping, stacked, or stacked-relative figure.

        :param active_categories: Ordered subset of categories to render.
        :type active_categories: list[str]
        :param categories: Full category list (used for color index lookup).
        :type categories: list[str]
        :param line_dfs: Per-category DataFrames.
        :type line_dfs: dict[str, pl.DataFrame]
        :param line_colors: Per-category color sequence.
        :type line_colors: HSLColorSequence
        :param alias: Date axis column name.
        :type alias: str
        :param value_range: Y-axis range applied when applicable.
        :type value_range: tuple[float, float]
        :return: Standard line figure.
        :rtype: go.Figure
        """
        traces = []
        for category in active_categories:
            i = categories.index(category)
            r, g, b = _hex_to_rgb(line_colors[i].to_css_hex())
            if self._placement == LinePlacement.OVERLAPPING:
                extra: dict[str, Any] = {}
                if self._fill_opacity > 0:
                    extra = dict(
                        fill="tozeroy",
                        fillcolor=f"rgba({r},{g},{b},{self._fill_opacity})",
                    )
            elif self._placement == LinePlacement.STACKED:
                extra = dict(
                    stackgroup="one",
                    fillcolor=f"rgba({r},{g},{b},{self._fill_opacity})",
                )
            else:
                extra = dict(
                    stackgroup="one",
                    groupnorm="percent",
                    fillcolor=f"rgba({r},{g},{b},{self._fill_opacity})",
                )
            y_data = line_dfs[category][self._aggregation.name]
            if self._placement != LinePlacement.OVERLAPPING:
                mask = y_data.is_not_null()
                x_data = line_dfs[category][alias].filter(mask)
                y_data = y_data.filter(mask)
            else:
                x_data = line_dfs[category][alias]
            traces.append(
                go.Scatter(
                    x=x_data,
                    y=y_data,
                    mode="lines",
                    line=dict(color=line_colors[i].to_css_hex(), width=self._line_width, shape=self._interpolation.value),
                    connectgaps=True,
                    name=str(category),
                    **extra,
                )
            )
        fig = go.Figure(data=traces)
        if self._placement == LinePlacement.OVERLAPPING:
            fig.update_yaxes(range=value_range)
        elif self._min_value is not None or self._max_value is not None:
            fig.update_yaxes(range=value_range)
        return fig

    def _configure_line_layout(
        self,
        fig: go.Figure,
        active_categories: list[str],
        categories: list[str],
        line_colors: HSLColorSequence,
        alias: str,
        effective_date: str,
        effective_value: str,
    ) -> None:
        """
        Apply axis titles, hover mode, legend, and ridgeline annotations.

        :param fig: Plotly figure to update.
        :type fig: go.Figure
        :param active_categories: Ordered subset of categories rendered.
        :type active_categories: list[str]
        :param categories: Full category list (used for color index lookup).
        :type categories: list[str]
        :param line_colors: Per-category color sequence.
        :type line_colors: HSLColorSequence
        :param alias: Date axis column name (unused for standard layout but kept for consistency).
        :type alias: str
        :param effective_date: Resolved date axis label.
        :type effective_date: str
        :param effective_value: Resolved value axis label.
        :type effective_value: str
        """
        if self._placement == LinePlacement.RIDGELINE:
            n_rows = len(active_categories)
            fig.update_xaxes(title_text=effective_date, row=n_rows, col=1)
            fig.update_xaxes(showgrid=False)
            fig.update_yaxes(showgrid=False, showticklabels=False)
            fig.update_layout(hovermode="x", showlegend=False)
            for idx, category in enumerate(active_categories):
                i = categories.index(category)
                axis_key = "yaxis" if idx == 0 else f"yaxis{idx + 1}"
                domain = fig.layout[axis_key].domain
                y_center = (domain[0] + domain[1]) / 2
                fig.add_annotation(
                    text=str(category),
                    xref="paper",
                    yref="paper",
                    x=1,
                    y=y_center,
                    xanchor="left",
                    yanchor="middle",
                    showarrow=False,
                    font=dict(color=line_colors[i].to_css_hex()),
                )
            top_domain = fig.layout["yaxis"].domain[1]
            last_axis_key = "yaxis" if n_rows == 1 else f"yaxis{n_rows}"
            bottom_domain = fig.layout[last_axis_key].domain[0]
            y_all_center = (top_domain + bottom_domain) / 2
            fig.add_annotation(
                text=effective_value,
                xref="paper",
                yref="paper",
                x=-0.07,
                y=y_all_center,
                xanchor="center",
                yanchor="middle",
                textangle=-90,
                showarrow=False,
            )
        else:
            fig.update_layout(
                xaxis_title=effective_date,
                yaxis_title=effective_value,
                hovermode="x",
                xaxis=dict(showgrid=False),
            )
            if self._explicit_category_variable is not None:
                fig.update_layout(legend_title_text=self._category_label or self._category_variable)

    def _prepare_plot_data(
        self,
        dfs: dict[str, pl.DataFrame],
        tr: Translations,
        theme: Theme,
    ) -> dict[str, Any]:
        df = dfs["main"]
        line_dfs, categories, alias = self._build_line_dataframes(df, tr)
        categories_df = pl.DataFrame({self._category_variable: categories})
        line_colors = self._compute_line_colors(categories_df, theme)
        value_range = self._compute_value_range(df)
        active_categories = self._sort_active_categories([c for c in categories if c in line_dfs], line_dfs)
        return {
            "categories": active_categories,
            "colors": [line_colors[categories.index(c)].to_css_hex() for c in active_categories],
            "_line_dfs": line_dfs,
            "_all_categories": categories,
            "_line_colors": line_colors,
            "_alias": alias,
            "_value_range": value_range,
        }

    def _render_from_payload(self, payload: dict[str, Any]) -> go.Figure:
        active_categories = payload["categories"]
        categories = payload["_all_categories"]
        line_dfs = payload["_line_dfs"]
        line_colors = payload["_line_colors"]
        alias = payload["_alias"]
        value_range = payload["_value_range"]
        if self._placement == LinePlacement.RIDGELINE:
            fig = self._build_ridgeline_figure(active_categories, categories, line_dfs, line_colors, alias, value_range)
        else:
            fig = self._build_standard_figure(active_categories, categories, line_dfs, line_colors, alias, value_range)
        effective_title = payload["title"] or (f"Line chart of {self._aggregation.name} by {self._date_variable}")
        effective_date = self._date_label or self._date_variable
        effective_value = self._value_label or self._append_unit(self._aggregation.default_label, self._aggregation.field)
        self._configure_line_layout(fig, active_categories, categories, line_colors, alias, effective_date, effective_value)
        self._apply_title(fig, effective_title, payload.get("subtitle"))
        return fig

    @classmethod
    def supported_color_selectors(cls) -> list[type[ColorSelectorStrategy]]:
        return [
            ColorSelectorStrategy,
            ColorSelectByCategoryName,
        ]

    @classmethod
    def supported_color_assigners(cls) -> list[type[ColorAssignerStrategy]]:
        return [
            ColorAssignerByCategory,
        ]


__all__ = [
    "LineChartStrategy",
]



"""
Grouped bar chart figure strategy implementation.
"""

from typing import Any

import plotly.express as px
import plotly.graph_objects as go
import polars as pl
from babel.support import Translations
from fields_metadata import FieldMetadata
from polychromos.palette import HSLColorSequence, Palette

from orama.color import (
    ColorAssignerByGroup,
    ColorAssignerStrategy,
    ColorRule,
    ColorSelectByCategoryName,
    ColorSelectorStrategy,
)
from orama.options import (
    BooleanOption,
    ChartOptionsMap,
)
from orama.query import QueryParams
from orama.strategies.bar import AbstractBarChartStrategy
from orama.theme import Theme
from orama.variables import (
    Aggregation,
    CategoricalVariable,
    ChartVariablesMap,
    NumericalAggregationVariable,
)


class GroupedBarChartStrategy(AbstractBarChartStrategy):
    """
    Grouped bar chart figure strategy.
    """

    _variables: ChartVariablesMap = {
        "category": CategoricalVariable(
            name="Category",
            description=(
                "Categorical variable(s) for the bar chart. "
                "Each category corresponds to a bar. "
                "If several variables are provided, they will be concatenated in a cartesian "
                "product with all of their values to form the final categories."
            ),
            map_to="category_variables",
            multicategorical=True,
        ),
        "value": NumericalAggregationVariable(
            name="Value",
            description=(
                "An aggregation of a numerical variable to produce the final value of each "
                "category. This value is proportional to the length of each bar."
            ),
            map_to="aggregation",
        ),
        "grouping": CategoricalVariable(
            name="Grouping",
            description=(
                "Categorical variable for grouping the bars. "
                "Categories are split in several groups, one per each value of this variable."
            ),
            map_to="group_variable",
            multicategorical=False,
        ),
    }
    _options: ChartOptionsMap = {
        "stacked": BooleanOption(
            name="Stacked bars",
            description=(
                "When enabled, group bars are stacked on top of each other. When disabled, group bars are rendered side by side."
            ),
            map_to="stacked",
            default_value=False,
        ),
        "relative": BooleanOption(
            name="Relative values",
            description=(
                "When enabled, bar values are shown as a percentage of the total value of their "
                "category. Only available when stacked is also enabled."
            ),
            map_to="relative",
            default_value=False,
        ),
        "horizontal": BooleanOption(
            name="Horizontal bars",
            description="When enabled, bars are rendered horizontally instead of vertically.",
            map_to="horizontal",
            default_value=False,
        ),
    }

    def __init__(
        self,
        fields_metadata: dict[str, FieldMetadata],
        category_variables: list[str],
        group_variable: str,
        aggregation: Aggregation,
        stacked: bool = False,
        relative: bool = False,
        horizontal: bool = False,
        color_rules: list[ColorRule] | None = None,
        title: str | None = None,
        subtitle: str | None = None,
        caption: str | None = None,
        description: str | None = None,
        storytelling_context: str | None = None,
        category_label: str | None = None,
        value_label: str | None = None,
        group_label: str | None = None,
    ) -> None:
        super().__init__(
            fields_metadata=fields_metadata,
            category_variables=category_variables,
            aggregation=aggregation,
            color_rules=color_rules,
            title=title,
            subtitle=subtitle,
            caption=caption,
            description=description,
            storytelling_context=storytelling_context,
            category_label=category_label,
            value_label=value_label,
        )

        self._variables["grouping"].validate_fields(fields_metadata, group_variable)

        self._options["stacked"].validate(stacked)
        self._options["relative"].validate(relative)
        if relative and not stacked:
            raise ValueError("Relative option can only be enabled when stacked is also enabled")
        self._options["horizontal"].validate(horizontal)

        self._group_variable = group_variable
        self._stacked = stacked
        self._relative = relative
        self._horizontal = horizontal
        self._group_label = group_label

    @classmethod
    def name(cls) -> str:
        return "Grouped Bar Chart"

    @classmethod
    def description(cls) -> str:
        return (
            "A grouped bar chart presents categorical data with rectangular bars with heights or "
            "lengths proportional to the values that they represent. The bars can be plotted "
            "vertically or horizontally. Categories are split in several groups, one per each "
            "value of a grouping variable. These charts add an extra categorical dimension to "
            "standard bar charts."
        )

    def get_query_params(self) -> dict[str, QueryParams]:
        query_params = super().get_query_params()
        query_params["main"].group_fields.append(self._group_variable)
        return query_params

    def _get_schema_fields(self) -> dict[str, type[pl.DataType]]:
        fields = super()._get_schema_fields()
        fields[self._group_variable] = self._infer_column_type(
            self._fields_metadata,
            self._group_variable,
        )
        return fields

    def _prepare_grouped_df(
        self,
        df: pl.DataFrame,
        tr: Translations,
    ) -> tuple[pl.DataFrame, str, str, list[str]]:
        """
        Sort/translate group and category variables, concatenate categories, compute category totals,
        and optionally normalise values for relative display.

        :param df: Raw DataFrame from the query result.
        :type df: pl.DataFrame
        :param tr: Translations for localizing labels.
        :type tr: Translations
        :return: Processed DataFrame, category alias, effective value column name, and sorted group list.
        :rtype: tuple[pl.DataFrame, str, str, list[str]]
        """
        df = self._sort_and_translate(
            df,
            self._group_variable,
            self._fields_metadata[self._group_variable],
            tr,
            sort_asc=True,
        )
        for category_variable in reversed(self._category_variables):
            df = self._sort_and_translate(
                df,
                category_variable,
                self._fields_metadata[category_variable],
                tr,
                sort_asc=True,
            )
        df, alias = self._concatenate_variables(df, self._category_variables)
        value_var = self._aggregation.name
        category_total_var = "category_total"
        total_per_category = df.group_by(alias).agg(pl.sum(value_var).alias(category_total_var))
        df = df.join(total_per_category, on=alias, how="left")
        relative = self._relative and self._stacked
        if relative and df.height > 0:
            new_value_var = f"Relative {value_var}"
            df = df.with_columns((pl.col(value_var) / pl.col(category_total_var)).alias(new_value_var))
            value_var = new_value_var
        sorted_groups = df[self._group_variable].unique(maintain_order=True).to_list()
        return df, alias, value_var, sorted_groups

    def _compute_group_color_map(
        self,
        sorted_groups: list[str],
        theme: Theme,
    ) -> dict[str, str] | None:
        """
        Build a ``color_discrete_map`` dict from color rules, or return ``None`` for the default palette.

        :param sorted_groups: Ordered list of group values.
        :type sorted_groups: list[str]
        :param theme: Theme providing the base color.
        :type theme: Theme
        :return: Hex color map keyed by group value, or ``None``.
        :rtype: dict[str, str] | None
        """
        if not self._color_rules:
            return None
        group_df = pl.DataFrame({self._group_variable: sorted_groups})
        n_groups = len(sorted_groups)
        group_colors: HSLColorSequence = [theme.get_color("primary_color_normal")] * n_groups
        for color_rule in self._color_rules:
            selector = color_rule.selector.selector(
                group_df,
                self._group_variable,
                self._aggregation.name,
            )
            assigned_colors = color_rule.assigner.assign_colors(
                group_df,
                self._group_variable,
                self._aggregation.name,
                self._group_variable,
            )
            group_colors = Palette.mix_color_sequences(
                group_colors,
                assigned_colors,
                selector,
                indexing=Palette.MixIndexing.BY_POSITION,
            )
        return {str(grp): group_colors[i].to_css_hex() for i, grp in enumerate(sorted_groups)}

    def _prepare_plot_data(
        self,
        dfs: dict[str, pl.DataFrame],
        tr: Translations,
        theme: Theme,
    ) -> dict[str, Any]:
        df, alias, value_var, sorted_groups = self._prepare_grouped_df(dfs["main"], tr)
        color_map = self._compute_group_color_map(sorted_groups, theme)
        return {
            "categories": df[alias].unique(maintain_order=True).to_list(),
            "groups": sorted_groups,
            "color_map": color_map,
            "_df": df,
            "_alias": alias,
            "_value_var": value_var,
        }

    def _render_from_payload(self, payload: dict[str, Any]) -> go.Figure:
        df = payload["_df"]
        alias = payload["_alias"]
        value_var = payload["_value_var"]
        color_map = payload["color_map"]
        relative = self._relative and self._stacked
        x_var = alias
        y_var = value_var
        if self._horizontal:
            x_var, y_var = y_var, x_var
        extra_hover_data: list[str] = ["category_total"]
        if relative:
            extra_hover_data.append(self._aggregation.name)
        bar_kwargs: dict[str, Any] = dict(
            barmode="stack" if self._stacked else "group",
            orientation="h" if self._horizontal else "v",
            hover_data=extra_hover_data,
        )
        if color_map is not None:
            fig = px.bar(
                df.to_pandas(),
                x=x_var,
                y=y_var,
                color=self._group_variable,
                color_discrete_map=color_map,
                **bar_kwargs,
            )
        else:
            fig = px.bar(
                df.to_pandas(),
                x=x_var,
                y=y_var,
                color=self._group_variable,
                **bar_kwargs,
            )
        effective_title = payload["title"] or (
            f"Grouped bar chart of {self._aggregation.name} by {', '.join(self._category_variables)} and {self._group_variable}"
        )
        effective_category = self._category_label or ", ".join(self._category_variables)
        effective_value = self._value_label or self._append_unit(self._aggregation.default_label, self._aggregation.field)
        effective_group = self._group_label or self._group_variable
        if self._horizontal:
            fig.update_layout(xaxis_title=effective_value, yaxis_title=effective_category)
        else:
            fig.update_layout(xaxis_title=effective_category, yaxis_title=effective_value)
        fig.update_layout(legend_title_text=effective_group)
        if relative:
            if self._horizontal:
                fig.update_xaxes(tickformat=".1%", range=[0, 1])
            else:
                fig.update_yaxes(tickformat=".1%", range=[0, 1])
        self._apply_title(fig, effective_title, payload.get("subtitle"))
        return fig

    @classmethod
    def supported_color_selectors(cls) -> list[type[ColorSelectorStrategy]]:
        return [
            ColorSelectorStrategy,
            ColorSelectByCategoryName,
        ]

    @classmethod
    def supported_color_assigners(cls) -> list[type[ColorAssignerStrategy]]:
        return [
            ColorAssignerByGroup,
        ]


__all__ = [
    "GroupedBarChartStrategy",
]



"""
Tests for FigureStrategy._build_pre_payload() and build() callback flow.
"""

from typing import Any

import polars as pl

from orama.color import ColorAssignerByCategory, ColorSelectorStrategy
from orama.strategies.base import FigureStrategy, FigureView

_DFS: dict[str, pl.DataFrame] = {"main": pl.DataFrame()}


class ConcreteStrategy(FigureStrategy):
    """Minimal concrete strategy for testing build() and _build_pre_payload()."""

    @classmethod
    def supported_color_selectors(cls) -> list[type]:
        return [ColorSelectorStrategy]

    @classmethod
    def supported_color_assigners(cls) -> list[type]:
        return [ColorAssignerByCategory]

    def get_query_params(self) -> dict[str, Any]:
        return {}

    def _render_from_payload(self, payload: dict[str, Any]) -> Any:
        import plotly.graph_objects as go

        fig = go.Figure()
        if payload.get("title"):
            self._apply_title(fig, payload["title"], payload.get("subtitle"))
        return fig

    def _get_schema_fields(self) -> dict[str, type[pl.DataType]]:
        return {}


def _make(
    title: str | None = None,
    subtitle: str | None = None,
    caption: str | None = None,
    description: str | None = None,
    storytelling_context: str | None = None,
    pre_build_callback: Any = None,
    post_build_callback: Any = None,
) -> ConcreteStrategy:
    return ConcreteStrategy(
        fields_metadata={},
        title=title,
        subtitle=subtitle,
        caption=caption,
        description=description,
        storytelling_context=storytelling_context,
        pre_build_callback=pre_build_callback,
        post_build_callback=post_build_callback,
    )


def test_pre_payload_is_flat() -> None:
    s = _make(title="T", subtitle="S", caption="C", description="D")
    pre = s._build_pre_payload(_DFS)
    assert "metadata" not in pre
    assert pre["title"] == "T"
    assert pre["subtitle"] == "S"
    assert pre["caption"] == "C"
    assert pre["description"] == "D"


def test_pre_payload_strategy_key() -> None:
    s = _make()
    pre = s._build_pre_payload(_DFS)
    assert pre["strategy"] == "ConcreteStrategy"


def test_pre_payload_dfs_is_copy_of_input() -> None:
    df = pl.DataFrame({"x": [1]})
    s = _make()
    pre = s._build_pre_payload({"main": df})
    assert "main" in pre["dfs"]


def test_pre_payload_none_fields() -> None:
    s = _make()
    pre = s._build_pre_payload(_DFS)
    assert pre["title"] is None
    assert pre["description"] is None


def test_pre_payload_includes_storytelling_context() -> None:
    s = _make(storytelling_context="Costs spiked in Q3 due to supply constraints.")
    pre = s._build_pre_payload(_DFS)
    assert pre["storytelling_context"] == "Costs spiked in Q3 due to supply constraints."


def test_pre_payload_storytelling_context_default_none() -> None:
    s = _make()
    pre = s._build_pre_payload(_DFS)
    assert pre["storytelling_context"] is None


def test_base_metadata_includes_storytelling_context() -> None:
    s = _make(storytelling_context="Trend shows steady growth.")
    meta = s._build_base_metadata()
    assert "storytelling_context" in meta
    assert meta["storytelling_context"] == "Trend shows steady growth."


def test_pre_payload_color_rules_is_list() -> None:
    s = _make()
    pre = s._build_pre_payload(_DFS)
    assert "color_rules" in pre
    assert isinstance(pre["color_rules"], list)


def test_pre_payload_variables_is_empty_for_no_variable_strategy() -> None:
    s = _make()
    pre = s._build_pre_payload(_DFS)
    assert pre["variables"] == {}


def test_pre_payload_options_is_empty_for_no_option_strategy() -> None:
    s = _make()
    pre = s._build_pre_payload(_DFS)
    assert pre["options"] == {}


def test_build_no_callbacks_returns_figure_view(tr: Any, theme: Any) -> None:
    s = _make(caption="Cap", description="Desc", storytelling_context="Story here.")
    view = s.build(_DFS, tr, theme)
    assert isinstance(view, FigureView)
    assert view.caption == "Cap"
    assert view.description == "Desc"
    assert view.storytelling_context == "Story here."


def test_pre_callback_title_propagates_to_figure(tr: Any, theme: Any) -> None:
    def cb(payload: dict[str, Any]) -> None:
        payload["title"] = "Overridden"

    s = _make(title="Original", pre_build_callback=cb)
    view = s.build(_DFS, tr, theme)
    assert view.figure.layout.title.text == "Overridden"


def test_pre_callback_description_propagates_to_view(tr: Any, theme: Any) -> None:
    def cb(payload: dict[str, Any]) -> None:
        payload["description"] = "From callback"

    s = _make(description="Original", pre_build_callback=cb)
    view = s.build(_DFS, tr, theme)
    assert view.description == "From callback"


def test_pre_callback_caption_propagates_to_view(tr: Any, theme: Any) -> None:
    def cb(payload: dict[str, Any]) -> None:
        payload["caption"] = "New caption"

    s = _make(caption="Old caption", pre_build_callback=cb)
    view = s.build(_DFS, tr, theme)
    assert view.caption == "New caption"


def test_pre_callback_subtitle_propagates_to_figure(tr: Any, theme: Any) -> None:
    def cb(payload: dict[str, Any]) -> None:
        payload["title"] = "T"
        payload["subtitle"] = "Sub"

    s = _make(pre_build_callback=cb)
    view = s.build(_DFS, tr, theme)
    assert view.figure.layout.title.subtitle.text == "Sub"


def test_pre_callback_dfs_mutation_is_used(tr: Any, theme: Any) -> None:
    """Mutations to payload["dfs"] during the callback are visible in _render_from_payload."""
    received: dict[str, Any] = {}

    def cb(payload: dict[str, Any]) -> None:
        payload["dfs"]["main"] = pl.DataFrame({"injected": [99]})

    class CaptureDfs(ConcreteStrategy):
        """Captures the payload passed to _render_from_payload() for assertion."""

        def _render_from_payload(self, payload: dict[str, Any]) -> Any:
            received["dfs"] = payload["dfs"]
            return super()._render_from_payload(payload)

    s = CaptureDfs(fields_metadata={}, pre_build_callback=cb)
    s.build({"main": pl.DataFrame()}, tr, theme)
    assert "injected" in received["dfs"]["main"].columns


def test_pre_callback_storytelling_context_propagates_to_view(tr: Any, theme: Any) -> None:
    def cb(payload: dict[str, Any]) -> None:
        payload["storytelling_context"] = "From callback"

    s = _make(storytelling_context="Original", pre_build_callback=cb)
    view = s.build(_DFS, tr, theme)
    assert view.storytelling_context == "From callback"


def test_pre_callback_does_not_mutate_instance(tr: Any, theme: Any) -> None:
    def cb(payload: dict[str, Any]) -> None:
        payload["title"] = "Overridden"
        payload["description"] = "New desc"
        payload["storytelling_context"] = "Changed"

    s = _make(title="Original", description="Old", storytelling_context="Unchanged", pre_build_callback=cb)
    s.build(_DFS, tr, theme)
    assert s._title == "Original"
    assert s._description == "Old"
    assert s._storytelling_context == "Unchanged"


def test_post_callback_description_overrides_view(tr: Any, theme: Any) -> None:
    def cb(payload: dict[str, Any]) -> None:
        payload["description"] = "Post desc"

    s = _make(description="Pre desc", post_build_callback=cb)
    view = s.build(_DFS, tr, theme)
    assert view.description == "Post desc"


def test_post_callback_caption_overrides_view(tr: Any, theme: Any) -> None:
    def cb(payload: dict[str, Any]) -> None:
        payload["caption"] = "Post cap"

    s = _make(caption="Pre cap", post_build_callback=cb)
    view = s.build(_DFS, tr, theme)
    assert view.caption == "Post cap"


def test_both_callbacks_post_wins_over_pre(tr: Any, theme: Any) -> None:
    def pre(payload: dict[str, Any]) -> None:
        payload["description"] = "From pre"

    def post(payload: dict[str, Any]) -> None:
        payload["description"] = "From post"

    s = _make(pre_build_callback=pre, post_build_callback=post)
    view = s.build(_DFS, tr, theme)
    assert view.description == "From post"


def test_plot_uses_instance_title(tr: Any, theme: Any) -> None:
    s = _make(title="My Title", caption="Cap", description="Desc")
    view = s.plot(_DFS, tr, theme)
    assert view.figure.layout.title.text == "My Title"
    assert view.caption == "Cap"
    assert view.description == "Desc"



"""
Various enumerations used in orama.
"""

from enum import Enum, StrEnum


class SortOrder(int, Enum):
    """
    Enumeration of possible sort orders.
    """

    ASCENDING = 1
    DESCENDING = -1
    NONE = 0


class LinePlacement(StrEnum):
    """
    Enumeration of placement modes for line chart series.
    """

    OVERLAPPING = "overlapping"
    STACKED = "stacked"
    STACKED_RELATIVE = "stacked_relative"
    RIDGELINE = "ridgeline"


class LineInterpolation(StrEnum):
    """
    Enumeration of interpolation modes for line chart traces.
    """

    LINEAR = "linear"
    SPLINE = "spline"
    STEP_HV = "hv"
    STEP_VH = "vh"


class LineSortOrder(StrEnum):
    """
    Enumeration of sort orders for line chart series/categories.
    """

    NONE = "none"
    ALPHABETICAL_ASC = "alpha_asc"
    ALPHABETICAL_DESC = "alpha_desc"
    FIRST_VALUE_ASC = "first_asc"
    FIRST_VALUE_DESC = "first_desc"
    LAST_VALUE_ASC = "last_asc"
    LAST_VALUE_DESC = "last_desc"


class GeoProjection(StrEnum):
    """
    Enumeration of geo projection types for choropleth maps.
    """

    NATURAL_EARTH = "natural earth"
    MERCATOR = "mercator"
    ORTHOGRAPHIC = "orthographic"
    EQUIRECTANGULAR = "equirectangular"
    ROBINSON = "robinson"
    KAVRAYSKIY7 = "kavrayskiy7"


__all__ = [
    "SortOrder",
    "LinePlacement",
    "LineInterpolation",
    "LineSortOrder",
    "GeoProjection",
]



"""
Unit tests for GroupedBarChartStrategy helper methods.
"""

from typing import Any, cast
from unittest.mock import MagicMock

import polars as pl
import pytest
from fields_metadata import FieldMetadata
from polychromos import HSLColor

from orama.color import (
    ColorAssignerByGroup,
    ColorRule,
    ColorSelectorStrategy,
)
from orama.strategies.grouped_bar import GroupedBarChartStrategy
from orama.variables import Aggregation


def _cat_meta() -> MagicMock:
    m = MagicMock(spec=FieldMetadata)
    m.categorical = True
    m.numeric = False
    m.derived = False
    m.effective_type = str
    m.extra = {}
    return m


def _make_grouped_bar(**kwargs: Any) -> GroupedBarChartStrategy:
    fields = cast(dict[str, FieldMetadata], {"cat": _cat_meta(), "grp": _cat_meta()})
    return GroupedBarChartStrategy(
        fields_metadata=fields,
        category_variables=["cat"],
        group_variable="grp",
        aggregation=Aggregation(name="count", function="count"),
        **kwargs,
    )


@pytest.fixture
def base_df() -> pl.DataFrame:
    return pl.DataFrame(
        {
            "cat": ["A", "A", "B", "B"],
            "grp": ["X", "Y", "X", "Y"],
            "count": [10, 20, 30, 40],
        }
    )


def test_prepare_grouped_df_returns_correct_columns(
    base_df: pl.DataFrame,
    tr: Any,
) -> None:
    instance = _make_grouped_bar()
    df, alias, value_var, sorted_groups = instance._prepare_grouped_df(base_df, tr)
    assert alias in df.columns
    assert "grp" in df.columns
    assert "count" in df.columns
    assert "category_total" in df.columns


def test_prepare_grouped_df_relative_adds_new_value_col(
    base_df: pl.DataFrame,
    tr: Any,
) -> None:
    instance = _make_grouped_bar(stacked=True, relative=True)
    df, alias, value_var, sorted_groups = instance._prepare_grouped_df(base_df, tr)
    assert value_var == "Relative count"
    assert "Relative count" in df.columns


def test_compute_group_color_map_no_rules_returns_none(
    theme: Any,
) -> None:
    instance = _make_grouped_bar()
    result = instance._compute_group_color_map(["X", "Y"], theme)
    assert result is None


def test_compute_group_color_map_with_rules_returns_dict(
    anchor_colors: list[HSLColor],
    theme: Any,
) -> None:
    assigner = ColorAssignerByGroup(anchor_colors=anchor_colors)
    rule = ColorRule(assigner=assigner, selector=ColorSelectorStrategy())
    instance = _make_grouped_bar(color_rules=[rule])
    result = instance._compute_group_color_map(["X", "Y"], theme)
    assert isinstance(result, dict)
    assert set(result.keys()) == {"X", "Y"}
    for v in result.values():
        assert v.startswith("#")


def test_prepare_plot_data_returns_categories_groups_color_map(
    base_df: pl.DataFrame,
    tr: Any,
    theme: Any,
) -> None:
    instance = _make_grouped_bar()
    result = instance._prepare_plot_data({"main": base_df}, tr, theme)
    assert isinstance(result["categories"], list)
    assert isinstance(result["groups"], list)
    assert result["color_map"] is None


def test_prepare_plot_data_color_map_with_rules(
    base_df: pl.DataFrame,
    anchor_colors: list[HSLColor],
    tr: Any,
    theme: Any,
) -> None:
    assigner = ColorAssignerByGroup(anchor_colors=anchor_colors)
    rule = ColorRule(assigner=assigner, selector=ColorSelectorStrategy())
    instance = _make_grouped_bar(color_rules=[rule])
    result = instance._prepare_plot_data({"main": base_df}, tr, theme)
    assert isinstance(result["color_map"], dict)
    assert set(result["color_map"].keys()) == {"X", "Y"}
    for v in result["color_map"].values():
        assert v.startswith("#")
